from functools import lru_cache
from pathlib import Path
from typing import Iterator

import pytest
from rdflib import BNode
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create import create_project
from dsp_tools.commands.validate_data.api_connection import ApiConnection
from dsp_tools.commands.validate_data.models.input_problems import ContentRegexProblem
from dsp_tools.commands.validate_data.models.input_problems import DuplicateValueProblem
from dsp_tools.commands.validate_data.models.input_problems import FileValueNotAllowedProblem
from dsp_tools.commands.validate_data.models.input_problems import FileValueProblem
from dsp_tools.commands.validate_data.models.input_problems import GenericProblem
from dsp_tools.commands.validate_data.models.input_problems import LinkedResourceDoesNotExistProblem
from dsp_tools.commands.validate_data.models.input_problems import LinkTargetTypeMismatchProblem
from dsp_tools.commands.validate_data.models.input_problems import MaxCardinalityProblem
from dsp_tools.commands.validate_data.models.input_problems import MinCardinalityProblem
from dsp_tools.commands.validate_data.models.input_problems import NonExistentCardinalityProblem
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.input_problems import ValueTypeProblem
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.reformat_validaton_result import _extract_base_info_of_resource_results
from dsp_tools.commands.validate_data.reformat_validaton_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _check_for_unknown_resource_classes
from dsp_tools.commands.validate_data.validate_data import _get_parsed_graphs
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from test.e2e_validate_data.setup_testcontainers import get_containers

CREDS = ServerCredentials("root@example.com", "test", "http://0.0.0.0:3333")
LOCAL_API = "http://0.0.0.0:3333"
DONT_SAVE_GRAPHS = False


@lru_cache(maxsize=None)
@pytest.fixture
def _create_project_generic() -> Iterator[None]:
    with get_containers():
        success = create_project(Path("testdata/validate-data/generic/project.json"), CREDS)
        assert success
        yield


@lru_cache(maxsize=None)
@pytest.fixture
def api_con() -> ApiConnection:
    return ApiConnection(LOCAL_API)


@lru_cache(maxsize=None)
@pytest.fixture
def cardinality_correct(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/cardinality_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def unknown_classes_graphs(_create_project_generic: Iterator[None], api_con: ApiConnection) -> RDFGraphs:
    file = Path("testdata/validate-data/generic/unknown_classes.xml")
    return _get_parsed_graphs(api_con, file)


@lru_cache(maxsize=None)
@pytest.fixture
def cardinality_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/cardinality_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def content_correct(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/content_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def content_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/content_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def every_combination_once(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/every_combination_once.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def minimal_correct(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/minimal_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def value_type_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/value_type_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def unique_value_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/unique_value_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def file_value_correct(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/file_value_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def file_value_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/file_value_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def dsp_inbuilt_correct(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def dsp_inbuilt_violation(_create_project_generic: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/generic/dsp_inbuilt_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def _create_project_special() -> Iterator[None]:
    with get_containers():
        success = create_project(
            Path("testdata/validate-data/special_characters/project_special_characters.json"), CREDS
        )
        assert success
        yield


@lru_cache(maxsize=None)
@pytest.fixture
def special_characters_correct(
    _create_project_special: Iterator[None], api_con: ApiConnection
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/special_characters/special_characters_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def special_characters_violation(
    _create_project_special: Iterator[None], api_con: ApiConnection
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/special_characters/special_characters_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def _create_project_inheritance() -> Iterator[None]:
    with get_containers():
        success = create_project(Path("testdata/validate-data/inheritance/project_inheritance.json"), CREDS)
        assert success
        yield


@lru_cache(maxsize=None)
@pytest.fixture
def inheritance_correct(_create_project_inheritance: Iterator[None], api_con: ApiConnection) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/inheritance/inheritance_correct.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


@lru_cache(maxsize=None)
@pytest.fixture
def inheritance_violation(
    _create_project_inheritance: Iterator[None], api_con: ApiConnection
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/inheritance/inheritance_violation.xml")
    graphs = _get_parsed_graphs(api_con, file)
    return _get_validation_result(graphs, api_con, file, DONT_SAVE_GRAPHS)


def test_extract_identifiers_of_resource_results(every_combination_once: ValidationReportGraphs) -> None:
    report_and_onto = every_combination_once.validation_graph + every_combination_once.onto_graph
    data_and_onto = every_combination_once.data_graph + every_combination_once.onto_graph
    result = _extract_base_info_of_resource_results(report_and_onto, data_and_onto)
    result_sorted = sorted(result, key=lambda x: str(x.resource_iri))
    expected_iris = [
        (URIRef("http://data/empty_label"), None),
        (URIRef("http://data/geoname_not_number"), BNode),
        (URIRef("http://data/id_card_one"), None),
        (URIRef("http://data/id_closed_constraint"), None),
        (URIRef("http://data/id_max_card"), None),
        (URIRef("http://data/id_missing_file_value"), None),
        (URIRef("http://data/id_simpletext"), BNode),
        (URIRef("http://data/id_uri"), BNode),
        (URIRef("http://data/identical_values"), None),
        (URIRef("http://data/link_target_non_existent"), BNode),
        (URIRef("http://data/link_target_wrong_class"), BNode),
        (URIRef("http://data/list_node_non_existent"), BNode),
    ]
    assert len(result) == len(expected_iris)
    for result_info, expected_iri in zip(result_sorted, expected_iris):
        assert result_info.resource_iri == expected_iri[0]
        if expected_iri[1] is None:
            assert not result_info.detail
        else:
            detail_base_info = result_info.detail
            assert isinstance(detail_base_info, DetailBaseInfo)
            assert isinstance(detail_base_info.detail_bn, expected_iri[1])


class TestCheckConforms:
    def test_cardinality_correct(self, cardinality_correct: ValidationReportGraphs) -> None:
        assert cardinality_correct.conforms

    def test_cardinality_violation(self, cardinality_violation: ValidationReportGraphs) -> None:
        assert not cardinality_violation.conforms

    def test_content_correct(self, content_correct: ValidationReportGraphs) -> None:
        assert content_correct.conforms

    def test_content_violation(self, content_violation: ValidationReportGraphs) -> None:
        assert not content_violation.conforms

    def test_every_combination_once(self, every_combination_once: ValidationReportGraphs) -> None:
        assert not every_combination_once.conforms

    def test_minimal_correct(self, minimal_correct: ValidationReportGraphs) -> None:
        assert minimal_correct.conforms

    def test_value_type_violation(self, value_type_violation: ValidationReportGraphs) -> None:
        assert not value_type_violation.conforms

    def test_unique_value_violation(self, unique_value_violation: ValidationReportGraphs) -> None:
        assert not unique_value_violation.conforms

    def test_file_value_correct(self, file_value_correct: ValidationReportGraphs) -> None:
        assert file_value_correct.conforms

    def test_file_value_cardinality_violation(self, file_value_violation: ValidationReportGraphs) -> None:
        assert not file_value_violation.conforms

    def test_dsp_inbuilt_correct(self, dsp_inbuilt_correct: ValidationReportGraphs) -> None:
        assert dsp_inbuilt_correct.conforms

    def test_dsp_inbuilt_violation(self, dsp_inbuilt_violation: ValidationReportGraphs) -> None:
        assert not dsp_inbuilt_violation.conforms

    def test_special_characters_correct(self, special_characters_correct: ValidationReportGraphs) -> None:
        assert special_characters_correct.conforms

    def test_special_characters_violation(self, special_characters_violation: ValidationReportGraphs) -> None:
        assert not special_characters_violation.conforms

    def test_inheritance_correct(self, inheritance_correct: ValidationReportGraphs) -> None:
        assert inheritance_correct.conforms

    def test_inheritance_violation(self, inheritance_violation: ValidationReportGraphs) -> None:
        assert not inheritance_violation.conforms


class TestReformatValidationGraph:
    def test_reformat_cardinality_violation(self, cardinality_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(cardinality_violation)
        expected_info_tuples = [
            (MinCardinalityProblem, "id_card_one"),
            (NonExistentCardinalityProblem, "id_closed_constraint"),
            (MaxCardinalityProblem, "id_max_card"),
            (MinCardinalityProblem, "id_min_card"),
            (NonExistentCardinalityProblem, "super_prop_no_card"),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert isinstance(one_result, expected_info[0])
            assert one_result.res_id == expected_info[1]

    def test_reformat_value_type_violation(self, value_type_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(value_type_violation)
        assert not result.unexpected_results
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        expected_info_tuples = [
            ("id_bool", "BooleanValue", "onto:testBoolean"),
            ("id_color", "ColorValue", "onto:testColor"),
            ("id_date", "DateValue", "onto:testSubDate1"),
            ("id_decimal", "DecimalValue", "onto:testDecimalSimpleText"),
            ("id_geoname", "GeonameValue", "onto:testGeoname"),
            ("id_integer", "IntValue", "onto:testIntegerSimpleText"),
            ("id_link", "LinkValue", "onto:testHasLinkTo"),
            ("id_list", "ListValue", "onto:testListProp"),
            ("id_richtext", "TextValue with formatting", "onto:testRichtext"),
            ("id_simpletext", "TextValue without formatting", "onto:testTextarea"),
            ("id_time", "TimeValue", "onto:testTimeValue"),
            ("id_uri", "UriValue", "onto:testUriValue"),
            ("is_link_should_be_integer", "IntValue", "onto:testIntegerSpinbox"),
            ("is_link_should_be_text", "TextValue without formatting", "onto:testTextarea"),
        ]
        assert len(result.problems) == len(expected_info_tuples)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert isinstance(one_result, ValueTypeProblem)
            assert one_result.res_id == expected_info[0]
            assert one_result.expected_type == expected_info[1]
            assert one_result.prop_name == expected_info[2]

    def test_reformat_content_violation(self, content_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(content_violation)
        assert not result.unexpected_results
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        expected_info_tuples = [
            ("empty_label", "rdfs:label", "The label must be a non-empty string"),
            ("empty_text_rich", "onto:testRichtext", "The value must be a non-empty string"),
            ("empty_text_simple", "onto:testTextarea", "The value must be a non-empty string"),
            ("geoname_not_number", "onto:testGeoname", "The value must be a valid geoname code"),
            ("link_target_non_existent", "onto:testHasLinkTo", "other"),
            ("link_target_wrong_class", "onto:testHasLinkToCardOneResource", "id_9_target"),
            (
                "list_name_attrib_empty",
                "onto:testListProp",
                "The list that should be used with this property is: firstList.",
            ),
            (
                "list_name_non_existent",
                "onto:testListProp",
                "The list that should be used with this property is: firstList.",
            ),
            ("list_node_non_existent", "onto:testListProp", "Unknown list node for list: firstList."),
            ("text_only_whitespace_simple", "onto:testTextarea", "The value must be a non-empty string"),
        ]
        assert len(result.problems) == len(expected_info_tuples)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert one_result.res_id == expected_info[0]
            assert one_result.prop_name == expected_info[1]
            if isinstance(one_result, ContentRegexProblem):
                assert one_result.expected_format == expected_info[2]
            elif isinstance(one_result, GenericProblem):
                assert one_result.results_message == expected_info[2]
            elif isinstance(one_result, LinkTargetTypeMismatchProblem):
                assert one_result.link_target_id == expected_info[2]
            elif isinstance(one_result, LinkedResourceDoesNotExistProblem):
                assert one_result.link_target_id == expected_info[2]
            else:
                assert isinstance(one_result, LinkedResourceDoesNotExistProblem)

    def test_reformat_every_constraint_once(self, every_combination_once: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(every_combination_once)
        expected_info_tuples = [
            ("empty_label", ContentRegexProblem),
            ("geoname_not_number", ContentRegexProblem),
            ("id_card_one", MinCardinalityProblem),
            ("id_closed_constraint", NonExistentCardinalityProblem),
            ("id_max_card", MaxCardinalityProblem),
            ("id_missing_file_value", FileValueProblem),
            ("id_simpletext", ValueTypeProblem),
            ("id_uri", ValueTypeProblem),
            ("identical_values", DuplicateValueProblem),
            ("link_target_non_existent", LinkedResourceDoesNotExistProblem),
            ("link_target_wrong_class", LinkTargetTypeMismatchProblem),
            ("list_node_non_existent", GenericProblem),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert one_result.res_id == expected_info[0]
            assert isinstance(one_result, expected_info[1])

    def test_reformat_unique_value_violation(self, unique_value_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(unique_value_violation)
        expected_ids = [
            "identical_values_LinkValue",
            "identical_values_listNode",
            "identical_values_valueAsString",
            "identical_values_valueHas",
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_ids)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_id in zip(sorted_problems, expected_ids):
            assert isinstance(one_result, DuplicateValueProblem)
            assert one_result.res_id == expected_id

    def test_reformat_file_value_violation(self, file_value_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(file_value_violation)
        expected_info_tuples = [
            ("id_archive_missing", FileValueProblem),
            ("id_archive_unknown", FileValueProblem),
            ("id_audio_missing", FileValueProblem),
            ("id_audio_unknown", FileValueProblem),
            ("id_document_missing", FileValueProblem),
            ("id_document_unknown", FileValueProblem),
            ("id_resource_without_representation", FileValueNotAllowedProblem),
            ("id_still_image_missing", FileValueProblem),
            ("id_still_image_unknown", FileValueProblem),
            ("id_text_missing", FileValueProblem),
            ("id_text_unknown", FileValueProblem),
            ("id_video_missing", FileValueProblem),
            ("id_video_unknown", FileValueProblem),
            ("id_wrong_file_type", FileValueProblem),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert isinstance(one_result, expected_info[1])
            assert one_result.res_id == expected_info[0]

    def test_reformat_dsp_inbuilt_violation(self, dsp_inbuilt_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(dsp_inbuilt_violation)
        expected_info_tuples = [
            ("link_obj_target_non_existent", LinkedResourceDoesNotExistProblem),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_info_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected_info in zip(sorted_problems, expected_info_tuples):
            assert isinstance(one_result, expected_info[1])
            assert one_result.res_id == expected_info[0]

    def test_reformat_special_characters_violation(self, special_characters_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(special_characters_violation)
        expected_tuples = [
            ("node_backslash", "Unknown list node for list: list \\ ' space.", "other \\ backslash"),
            ("node_double_quote", "Unknown list node for list: list \\ ' space.", 'other double quote "'),
            ("node_single_quote", "Unknown list node for list: list \\ ' space.", "other single quote '"),
            ("non_ascii_latin_alphabet", "", ""),
            ("non_ascii_other_alphabet", "", ""),
            ("special_char", "", ""),
            ("wrong_list_name", "The list that should be used with this property is: list \\ ' space.", "other"),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_tuples)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for prblm, expected in zip(sorted_problems, expected_tuples):
            if isinstance(prblm, GenericProblem):
                assert prblm.res_id == expected[0]
                assert prblm.problem == expected[1]
                assert prblm.actual_content == expected[2]
            elif isinstance(prblm, ContentRegexProblem):
                assert prblm.res_id == expected[0]

    def test_reformat_inheritance_violation(self, inheritance_violation: ValidationReportGraphs) -> None:
        result = reformat_validation_graph(inheritance_violation)
        expected_results = [
            ("ResourceSubCls1", {"onto:hasText0"}),
            ("ResourceSubCls2", {"onto:hasTextSubProp1", "onto:hasText0"}),
            ("ResourceSubCls2", {"onto:hasTextSubProp1", "onto:hasText0"}),
            ("ResourceUnrelated", {"onto:hasText0"}),
        ]
        assert not result.unexpected_results
        assert len(result.problems) == len(expected_results)
        sorted_problems = sorted(result.problems, key=lambda x: x.res_id)
        for one_result, expected in zip(sorted_problems, expected_results):
            assert isinstance(one_result, NonExistentCardinalityProblem)
            assert one_result.res_id == expected[0]
            assert one_result.prop_name in expected[1]


def test_check_for_unknown_resource_classes(unknown_classes_graphs: RDFGraphs) -> None:
    result = _check_for_unknown_resource_classes(unknown_classes_graphs)
    assert isinstance(result, UnknownClassesInData)
    expected = {"onto:NonExisting", "unknown:ClassWithEverything", "unknownClass"}
    assert result.unknown_classes == expected


if __name__ == "__main__":
    pytest.main([__file__])



"""end to end tests for ontology class"""

import unittest

import pytest

from dsp_tools.commands.project.models.ontology import Ontology
from dsp_tools.models.datetimestamp import DateTimeStamp
from dsp_tools.utils.authentication_client_live import AuthenticationClientLive
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestOntology(unittest.TestCase):
    con: Connection
    test_project = "http://rdfh.ch/projects/0001"
    test_onto = "http://0.0.0.0:3333/ontology/0001/anything/v2"

    def setUp(self) -> None:
        """
        Creates a connection to DSP-API.
        For each test method, a new TestCase instance is created, so setUp() is executed before each test method.
        """
        auth = AuthenticationClientLive("http://0.0.0.0:3333", "root@example.com", "test")
        self.con = ConnectionLive("http://0.0.0.0:3333", auth)

    def tearDown(self) -> None:
        """
        Logs out from DSP-API.
        For each test method, a new TestCase instance is created, so tearDown() is executed after each test method.
        """
        self.con.logout()

    def test_Ontology(self) -> None:
        last_mod_date_str = "2017-12-19T15:23:42.166Z"
        onto = Ontology(
            con=self.con,
            project=self.test_project,
            name="test_onto_name",
            label="Test ontology label",
            lastModificationDate=last_mod_date_str,
        )
        self.assertEqual(onto.project, self.test_project)
        self.assertEqual(onto.name, "test_onto_name")
        self.assertEqual(onto.label, "Test ontology label")
        self.assertEqual(onto.lastModificationDate, DateTimeStamp(last_mod_date_str))

    def test_ontology_create(self) -> None:
        onto = Ontology(
            con=self.con,
            project=self.test_project,
            name="test_onto_create",
            label="Test ontology create label",
        ).create()

        self.assertIsNotNone(onto.iri)
        self.assertEqual(onto.iri, "http://0.0.0.0:3333/ontology/0001/" + "test_onto_create" + "/v2")
        self.assertEqual(onto.project, self.test_project)
        self.assertEqual(onto.name, "test_onto_create")
        self.assertEqual(onto.label, "Test ontology create label")
        self.assertIsNotNone(onto.lastModificationDate)

    def test_ontology_getProjectOntologies(self) -> None:
        onto_list = Ontology.getProjectOntologies(self.con, self.test_project)
        onto_list_ids = [lst.iri for lst in onto_list]
        self.assertIn("http://0.0.0.0:3333/ontology/0001/anything/v2", onto_list_ids)
        self.assertIn("http://0.0.0.0:3333/ontology/0001/test_onto_create/v2", onto_list_ids)


if __name__ == "__main__":
    pytest.main([__file__])



from lxml import etree
from pytest_unordered import unordered

from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import IIIFUriInfo
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLBitstream
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLProperty
from dsp_tools.commands.xmlupload.models.deserialise.deserialise_value import XMLValue
from dsp_tools.commands.xmlupload.models.deserialise.xmlresource import XMLResource
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue


def test_get_restype_knora_api() -> None:
    res = XMLResource._get_restype(etree.fromstring("""<resource restype="Region"/>"""), "rosetta")
    assert res == "knora-api:Region"


def test_get_restype_abbreviated() -> None:
    res = XMLResource._get_restype(etree.fromstring("""<resource restype=":TestThing"/>"""), "rosetta")
    assert res == "rosetta:TestThing"


def test_get_restype_full() -> None:
    res = XMLResource._get_restype(etree.fromstring("""<resource restype="rosetta:TestThing"/>"""), "rosetta")
    assert res == "rosetta:TestThing"


def test_get_properties_bitstream_only() -> None:
    string = "<resource><bitstream>testdata/bitstreams/test.tif</bitstream></resource>"
    bitstream_expected = XMLBitstream("testdata/bitstreams/test.tif")
    bitstream, iiif, props = XMLResource._get_properties(etree.fromstring(string), "rosetta")
    assert bitstream == bitstream_expected
    assert not iiif
    assert not props


def test_get_properties_iiif_only() -> None:
    iiif_uri = "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2"
    string = f"<resource><iiif-uri>{iiif_uri}</iiif-uri></resource>"
    iiif_expected = IIIFUriInfo(iiif_uri)
    bitstream, iiif, props = XMLResource._get_properties(etree.fromstring(string), "rosetta")
    assert not bitstream
    assert iiif == iiif_expected
    assert not props


def test_get_properties_of_normal_resource() -> None:
    string = """
        <resource>
            <boolean-prop name=":hasBoolean"><boolean>true</boolean></boolean-prop>
            <color-prop name=":hasColor"><color>#00ff00</color></color-prop>
            <date-prop name=":hasDate"><date>JULIAN:BCE:70:CE:1</date></date-prop>
            <decimal-prop name=":hasDecimal"><decimal>2.718</decimal></decimal-prop>
            <geometry-prop name="hasGeometry"><geometry>{...}</geometry></geometry-prop>
            <geoname-prop name=":hasGeoname"><geoname>5416656</geoname></geoname-prop>
            <integer-prop name=":hasInteger"><integer>4711</integer></integer-prop>
            <list-prop list="testlist" name=":hasListItem"><list>subnode</list></list-prop>
            <resptr-prop name=":hasLink"><resptr>test_thing_0</resptr></resptr-prop>
            <text-prop name=":hasSimpleText"><text encoding="utf8">foo</text></text-prop>
            <time-prop name=":hasTime"><time>2019-10-23T13:45:12.01-14:00</time></time-prop>
            <uri-prop name=":hasUri"><uri>https://dasch.swiss</uri></uri-prop>
        </resource>
    """
    props_expected = [
        XMLProperty("rosetta:hasBoolean", "boolean", [XMLValue("true")]),
        XMLProperty("rosetta:hasColor", "color", [XMLValue("#00ff00")]),
        XMLProperty("rosetta:hasDate", "date", [XMLValue("JULIAN:BCE:70:CE:1")]),
        XMLProperty("rosetta:hasDecimal", "decimal", [XMLValue("2.718")]),
        XMLProperty("knora-api:hasGeometry", "geometry", [XMLValue("{...}")]),
        XMLProperty("rosetta:hasGeoname", "geoname", [XMLValue("5416656")]),
        XMLProperty("rosetta:hasInteger", "integer", [XMLValue("4711")]),
        XMLProperty("rosetta:hasListItem", "list", [XMLValue("testlist:subnode")]),
        XMLProperty("rosetta:hasLink", "resptr", [XMLValue("test_thing_0")]),
        XMLProperty("rosetta:hasSimpleText", "text", [XMLValue("foo")]),
        XMLProperty("rosetta:hasTime", "time", [XMLValue("2019-10-23T13:45:12.01-14:00")]),
        XMLProperty("rosetta:hasUri", "uri", [XMLValue("https://dasch.swiss")]),
    ]
    bitstream, iiif, props = XMLResource._get_properties(etree.fromstring(string), "rosetta")
    assert not bitstream
    assert not iiif
    assert props == unordered(props_expected)


def test_get_properties_of_audio_segment() -> None:
    string = """
        <audio-segment>
            <isAudioSegmentOf>audio_thing_1</isAudioSegmentOf>
            <hasSegmentBounds segment_start="10" segment_end="30"/>
            <hasTitle>Title</hasTitle>
            <hasComment>Comment</hasComment>
            <hasDescription>Description</hasDescription>
            <hasKeyword>Keyword</hasKeyword>
            <relatesTo>video_segment_0</relatesTo>
        </audio-segment>
    """
    props_expected = [
        XMLProperty("knora-api:isAudioSegmentOf", "resptr", [XMLValue("audio_thing_1")]),
        XMLProperty("knora-api:hasSegmentBounds", "interval", [XMLValue("10:30")]),
        XMLProperty("knora-api:hasTitle", "text", [XMLValue("Title")]),
        XMLProperty("knora-api:hasComment", "text", [XMLValue(FormattedTextValue("Comment"))]),
        XMLProperty("knora-api:hasDescription", "text", [XMLValue(FormattedTextValue("Description"))]),
        XMLProperty("knora-api:hasKeyword", "text", [XMLValue("Keyword")]),
        XMLProperty("knora-api:relatesTo", "resptr", [XMLValue("video_segment_0")]),
    ]
    bitstream, iiif, props = XMLResource._get_properties(etree.fromstring(string), "rosetta")
    assert not bitstream
    assert not iiif
    assert props == unordered(props_expected)


def test_get_properties_of_video_segment() -> None:
    string = """
        <video-segment>
            <isVideoSegmentOf>video_thing_1</isVideoSegmentOf>
            <hasSegmentBounds segment_start="10" segment_end="30"/>
            <hasTitle>Title</hasTitle>
            <hasComment>Comment</hasComment>
            <hasDescription>Description</hasDescription>
            <hasKeyword>Keyword</hasKeyword>
            <relatesTo>audio_thing_1</relatesTo>
        </video-segment>
    """
    props_expected = [
        XMLProperty("knora-api:isVideoSegmentOf", "resptr", [XMLValue("video_thing_1")]),
        XMLProperty("knora-api:hasSegmentBounds", "interval", [XMLValue("10:30")]),
        XMLProperty("knora-api:hasTitle", "text", [XMLValue("Title")]),
        XMLProperty("knora-api:hasComment", "text", [XMLValue(FormattedTextValue("Comment"))]),
        XMLProperty("knora-api:hasDescription", "text", [XMLValue(FormattedTextValue("Description"))]),
        XMLProperty("knora-api:hasKeyword", "text", [XMLValue("Keyword")]),
        XMLProperty("knora-api:relatesTo", "resptr", [XMLValue("audio_thing_1")]),
    ]
    bitstream, iiif, props = XMLResource._get_properties(etree.fromstring(string), "rosetta")
    assert not bitstream
    assert not iiif
    assert props == unordered(props_expected)


def test_group_props() -> None:
    comment_1 = XMLValue(FormattedTextValue("Comment 1"))
    comment_2 = XMLValue(FormattedTextValue("Comment 2"))
    desc_1 = XMLValue(FormattedTextValue("Description 1"))
    desc_2 = XMLValue(FormattedTextValue("Description 2"))
    kw_1 = XMLValue("Keyword 1")
    kw_2 = XMLValue("Keyword 2")
    link_1 = XMLValue("other_res_1")
    link_2 = XMLValue("other_res_2")
    ungrouped = [
        XMLProperty("knora-api:isAudioSegmentOf", "resptr", [XMLValue("audio_thing_1")]),
        XMLProperty("knora-api:hasSegmentBounds", "interval", [XMLValue("10:30")]),
        XMLProperty("knora-api:hasTitle", "text", [XMLValue("Title")]),
        XMLProperty("knora-api:hasComment", "text", [comment_1]),
        XMLProperty("knora-api:hasDescription", "text", [desc_1]),
        XMLProperty("knora-api:hasKeyword", "text", [kw_1]),
        XMLProperty("knora-api:relatesTo", "resptr", [link_1]),
        XMLProperty("knora-api:hasComment", "text", [comment_2]),
        XMLProperty("knora-api:hasDescription", "text", [desc_2]),
        XMLProperty("knora-api:hasKeyword", "text", [kw_2]),
        XMLProperty("knora-api:relatesTo", "resptr", [link_2]),
    ]
    expected = [
        XMLProperty("knora-api:isAudioSegmentOf", "resptr", [XMLValue("audio_thing_1")]),
        XMLProperty("knora-api:hasSegmentBounds", "interval", [XMLValue("10:30")]),
        XMLProperty("knora-api:hasTitle", "text", [XMLValue("Title")]),
        XMLProperty("knora-api:hasComment", "text", [comment_1, comment_2]),
        XMLProperty("knora-api:hasDescription", "text", [desc_1, desc_2]),
        XMLProperty("knora-api:hasKeyword", "text", [kw_1, kw_2]),
        XMLProperty("knora-api:relatesTo", "resptr", [link_1, link_2]),
    ]
    grouped = XMLResource._group_props(ungrouped)
    assert grouped == unordered(expected)


def test_get_props_with_links() -> None:
    link_1 = """<a class="salsah-link" href="IRI:test_thing_3:IRI">link</a>"""
    link_2 = """<a class="salsah-link" href="IRI:test_thing_4:IRI">link</a>"""
    string = f"""
        <resource label="label" id="id" restype=":foo">
            <boolean-prop name=":hasBoolean"><boolean>true</boolean></boolean-prop>
            <resptr-prop name=":hasLink">
                <resptr>test_thing_1</resptr>
            </resptr-prop>
            <resptr-prop name=":hasOtherLink">
                <resptr>test_thing_2</resptr>
            </resptr-prop>
            <text-prop name=":hasText1">
                <text encoding="xml">{link_1}</text>
            </text-prop>
            <text-prop name=":hasText2">
                <text encoding="xml">{link_2}</text>
            </text-prop>
            <uri-prop name=":hasUri"><uri>https://dasch.swiss</uri></uri-prop>
        </resource>
    """
    xml_resource = XMLResource.from_node(etree.fromstring(string), "rosetta")
    expected = [
        XMLProperty("rosetta:hasLink", "resptr", [XMLValue("test_thing_1")]),
        XMLProperty("rosetta:hasOtherLink", "resptr", [XMLValue("test_thing_2")]),
        XMLProperty("rosetta:hasText1", "text", [XMLValue(FormattedTextValue(link_1), resrefs={"test_thing_3"})]),
        XMLProperty("rosetta:hasText2", "text", [XMLValue(FormattedTextValue(link_2), resrefs={"test_thing_4"})]),
    ]
    res = xml_resource.get_props_with_links()
    assert res == unordered(expected)



"""end to end tests for resourceClass class"""

import unittest

import pytest

from dsp_tools.commands.project.models.ontology import Ontology
from dsp_tools.commands.project.models.resourceclass import ResourceClass
from dsp_tools.models.langstring import LangString
from dsp_tools.models.langstring import Languages
from dsp_tools.utils.authentication_client_live import AuthenticationClientLive
from dsp_tools.utils.connection import Connection
from dsp_tools.utils.connection_live import ConnectionLive

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestResourceClass(unittest.TestCase):
    test_project = "http://rdfh.ch/projects/0001"
    res_name = "res_class_name"
    res_label = LangString({Languages.EN: "Resource Class Label"})
    res_comment = LangString({Languages.EN: "This is a resource class for testing"})
    con: Connection

    def setUp(self) -> None:
        """
        Creates a connection to DSP-API.
        For each test method, a new TestCase instance is created, so setUp() is executed before each test method.
        """
        auth = AuthenticationClientLive("http://0.0.0.0:3333", "root@example.com", "test")
        self.con = ConnectionLive("http://0.0.0.0:3333", auth)

    def tearDown(self) -> None:
        """
        Logs out from DSP-API.
        For each test method, a new TestCase instance is created, so tearDown() is executed after each test method.
        """
        self.con.logout()

    def test_ResourceClass_create(self) -> None:
        onto = Ontology(
            con=self.con,
            project=self.test_project,
            name="test_onto",
            label="Test Ontology",
        ).create()

        last_modification_date_onto = onto.lastModificationDate

        # create test resource class
        _, res_class = ResourceClass(
            con=self.con,
            context=onto.context,
            name=self.res_name,
            ontology_id=onto.iri,
            label=self.res_label,
            comment=self.res_comment,
        ).create(last_modification_date_onto)

        self.assertIsNotNone(res_class.iri)
        self.assertEqual(res_class.name, self.res_name)
        self.assertEqual(res_class.label["en"], self.res_label["en"])
        self.assertEqual(res_class.comment["en"], self.res_comment["en"])

    def test_ResourceClass_update(self) -> None:
        onto = Ontology(
            con=self.con,
            project=self.test_project,
            name="test_onto_2",
            label="Test Ontology 2",
        ).create()

        last_modification_date = onto.lastModificationDate

        # create test resource class
        last_modification_date, res_class = ResourceClass(
            con=self.con,
            context=onto.context,
            name=self.res_name,
            ontology_id=onto.iri,
            label=self.res_label,
            comment=self.res_comment,
        ).create(last_modification_date)

        onto.lastModificationDate = last_modification_date

        self.assertIsNotNone(res_class.iri)


if __name__ == "__main__":
    pytest.main([__file__])



from rdflib.term import Node


def reformat_onto_iri(iri: Node) -> str:
    """Takes a rdflib Node and returns a prefixed IRI in string form."""
    iri_str = str(iri)
    if "http://www.w3.org/2000/01/rdf-schema#" in iri_str:
        return f'rdfs:{iri_str.split("#")[-1]}'
    onto = iri_str.split("/")[-2]
    ending = iri_str.split("#")[-1]
    if onto == "knora-api":
        return ending
    return f"{onto}:{ending}"


def reformat_data_iri(iri: Node) -> str:
    """Takes a rdflib Node with in the data namespace and returns only the suffix."""
    return str(iri).replace("http://data/", "")



import json
from pathlib import Path
from typing import Iterator

import pandas as pd
import pytest
from pytest_unordered import unordered

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.ingest_xmlupload.create_resources.upload_xml import ingest_xmlupload
from dsp_tools.commands.ingest_xmlupload.ingest_files.ingest_files import ingest_files
from dsp_tools.commands.ingest_xmlupload.upload_files.upload_files import upload_files
from dsp_tools.commands.project.create.project_create import create_project
from test.e2e.setup_testcontainers import SIPI_IMAGES
from test.e2e.setup_testcontainers import TMP_INGEST
from test.e2e.setup_testcontainers import get_containers

CREDS = ServerCredentials("root@example.com", "test", "http://0.0.0.0:3333")
CWD = Path("testdata/dsp-ingest-data/e2e-sample-project")
XML_FILE = Path("data.xml")
MULTIMEDIA_FILE_1 = Path("Bilder Projekt 2024/Côté gauche/Bild A (1).jpg")
MULTIMEDIA_FILE_2 = Path("Bilder Projekt 2024/Côté gauche/Dokument B (2).pdf")
SHORTCODE = "4125"
TMP_FOLDER = TMP_INGEST / "import" / SHORTCODE


@pytest.fixture
def mapping_file(monkeypatch: pytest.MonkeyPatch) -> Iterator[Path]:
    with monkeypatch.context() as m:
        m.chdir(CWD)
        mapping_file = Path(f"mapping-{SHORTCODE}.csv")
        yield mapping_file
        mapping_file.unlink(missing_ok=True)
        for id2iri_mapping in Path(".").glob("*id2iri_mapping*.json"):
            id2iri_mapping.unlink(missing_ok=True)


@pytest.fixture
def _create_project() -> Iterator[None]:
    with get_containers():
        success = create_project(Path("testdata/dsp-ingest-data/e2e-sample-project/project.json"), CREDS, verbose=True)
        assert success
        yield


@pytest.mark.usefixtures("_create_project")
def test_ingest_upload(mapping_file: Path) -> None:
    _test_upload_step()
    _test_ingest_step(mapping_file)
    _test_xmlupload_step()


def _test_upload_step() -> None:
    success = upload_files(XML_FILE, CREDS, Path(".").absolute())
    assert success
    assert {x.relative_to(TMP_FOLDER) for x in TMP_FOLDER.glob("**/*.*")} == {MULTIMEDIA_FILE_1, MULTIMEDIA_FILE_2}


def _test_ingest_step(mapping_file: Path) -> None:
    success = ingest_files(CREDS, SHORTCODE)
    assert success
    assert not {x.relative_to(TMP_FOLDER) for x in TMP_FOLDER.glob("**/*.*")}
    ingested_files = list((SIPI_IMAGES / SHORTCODE).glob("**/*.*"))
    ingested_files_ext = [f.suffixes for f in ingested_files]
    expected_ext = [[".info"], [".jpg", ".orig"], [".jpx"], [".info"], [".pdf", ".orig"], [".pdf"]]
    assert unordered(ingested_files_ext) == expected_ext

    df = pd.read_csv(mapping_file)
    assert df["original"].tolist() == unordered([str(MULTIMEDIA_FILE_1), str(MULTIMEDIA_FILE_2)])


def _test_xmlupload_step() -> None:
    success = ingest_xmlupload(XML_FILE, CREDS)
    assert success
    id2iri_file = list(Path.cwd().glob("*_id2iri_mapping_localhost.json"))[-1]  # choose the most recent one
    id2iri_mapping = json.loads(id2iri_file.read_text(encoding="utf-8"))
    assert sorted(id2iri_mapping.keys()) == ["resource_1", "resource_2", "resource_3"]
    assert all(x.startswith(f"http://rdfh.ch/{SHORTCODE}/") for x in id2iri_mapping.values())



import subprocess
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator

import docker
import regex
import requests
from docker.errors import NotFound
from docker.models.networks import Network
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

E2E_TESTDATA = Path("testdata/e2e").absolute()
SIPI_IMAGES = E2E_TESTDATA / "images"
TMP_SIPI = E2E_TESTDATA / "tmp-dsp-sipi"
TMP_INGEST = E2E_TESTDATA / "tmp-dsp-ingest"
INGEST_DB = E2E_TESTDATA / "ingest-db"


@dataclass(frozen=True)
class ImageVersions:
    fuseki: str
    sipi: str
    ingest: str
    api: str


@dataclass(frozen=True)
class Containers:
    fuseki: DockerContainer
    sipi: DockerContainer
    ingest: DockerContainer
    api: DockerContainer


@contextmanager
def get_test_network() -> Iterator[Network]:
    name = "dsp-tools-test-network"
    client = docker.client.from_env()
    try:
        network = client.networks.get(name)
    except NotFound:
        client.networks.create(name, internal=True, check_duplicate=True)
        network = client.networks.get(name)
    try:
        yield network
    finally:
        network.remove()
        client.close()  # type: ignore[no-untyped-call]  # incomplete stubs - remove this when the stubs are complete


@contextmanager
def get_containers() -> Iterator[Containers]:
    if subprocess.run("docker stats --no-stream".split(), check=False).returncode != 0:
        raise RuntimeError("Docker is not running properly")
    with get_test_network() as network:
        containers = _get_all_containers(network)
        try:
            yield containers
        finally:
            _stop_all_containers(containers)


def _get_all_containers(network: Network) -> Containers:
    versions = _get_image_versions()
    fuseki = _get_fuseki_container(network, versions.fuseki)
    sipi = _get_sipi_container(network, versions.sipi)
    ingest = _get_ingest_container(network, versions.ingest)
    api = _get_api_container(network, versions.api)
    containers = Containers(fuseki=fuseki, sipi=sipi, ingest=ingest, api=api)
    _print_containers_are_ready(containers)
    return containers


def _get_image_version(docker_compose_content: str, component: str) -> str:
    match = regex.search(rf"image: daschswiss/{component}:([^\n]+)", docker_compose_content)
    return match.group(1) if match else "latest"


def _get_api_image_version() -> str:
    docker_compose_content = Path("src/dsp_tools/resources/start-stack/docker-compose-validation.yml").read_text(
        encoding="utf-8"
    )
    match = regex.search(r"image: daschswiss/knora-api:([^\n]+)", docker_compose_content)
    return match.group(1) if match else "latest"


def _get_image_versions() -> ImageVersions:
    docker_compose_content = Path("src/dsp_tools/resources/start-stack/docker-compose.yml").read_text(encoding="utf-8")
    fuseki = _get_image_version(docker_compose_content, "apache-jena-fuseki")
    sipi = _get_image_version(docker_compose_content, "knora-sipi")
    ingest = _get_image_version(docker_compose_content, "dsp-ingest")
    api = _get_api_image_version()
    return ImageVersions(fuseki=fuseki, sipi=sipi, ingest=ingest, api=api)


def _get_fuseki_container(network: Network, version: str) -> DockerContainer:
    fuseki = (
        DockerContainer(f"daschswiss/apache-jena-fuseki:{version}")
        .with_name("db")
        .with_network(network)
        .with_bind_ports(3030, 3030)
        .with_env("ADMIN_PASSWORD", "test")
    )
    fuseki.start()
    wait_for_logs(fuseki, r"Server .+ Started .+ on port \d+$")
    print("Fuseki is ready")
    _create_data_set_and_admin_user()
    return fuseki


def _create_data_set_and_admin_user() -> None:
    repo_config = Path("testdata/e2e/repo_config.ttl").read_text(encoding="utf-8")
    files = {"file": ("file.ttl", repo_config, "text/turtle; charset=utf8")}
    if not requests.post("http://0.0.0.0:3030/$/datasets", files=files, auth=("admin", "test"), timeout=30).ok:
        raise RuntimeError("Fuseki did not create the dataset")
    print("Dataset created")

    admin_user_data = Path("testdata/e2e/admin_user_data.ttl").read_text(encoding="utf-8")
    url = "http://0.0.0.0:3030/knora-test/data?graph=http://www.knora.org/data/admin"
    files = {"file": ("file.ttl", admin_user_data, "text/turtle; charset: utf-8")}
    if not requests.post(url, files=files, auth=("admin", "test"), timeout=30).ok:
        raise RuntimeError("Fuseki did not create the admin user")
    print("Admin user created")


def _get_sipi_container(network: Network, version: str) -> DockerContainer:
    sipi = (
        DockerContainer(f"daschswiss/knora-sipi:{version}")
        .with_name("sipi")
        .with_network(network)
        .with_bind_ports(1024, 1024)
        .with_env("KNORA_WEBAPI_KNORA_API_EXTERNAL_HOST", "0.0.0.0")  # noqa: S104
        .with_env("KNORA_WEBAPI_KNORA_API_EXTERNAL_PORT", "3333")
        .with_command("--config=/sipi/config/sipi.docker-config.lua")
        .with_volume_mapping(TMP_SIPI, "/tmp", "rw")  # noqa: S108
        .with_volume_mapping(E2E_TESTDATA, "/sipi/config", "rw")
        .with_volume_mapping(SIPI_IMAGES, "/sipi/images", "rw")
    )
    sipi.start()
    wait_for_logs(sipi, "Sipi: Server listening on HTTP port 1024")
    print("Sipi is ready")
    return sipi


def _get_ingest_container(network: Network, version: str) -> DockerContainer:
    ingest = (
        DockerContainer(f"daschswiss/dsp-ingest:{version}")
        .with_name("ingest")
        .with_network(network)
        .with_bind_ports(3340, 3340)
        .with_env("STORAGE_ASSET_DIR", "/opt/images")
        .with_env("STORAGE_TEMP_DIR", "/opt/temp")
        .with_env("JWT_AUDIENCE", "http://localhost:3340")
        .with_env("JWT_ISSUER", "0.0.0.0:3333")
        .with_env("JWT_SECRET", "UP 4888, nice 4-8-4 steam engine")
        .with_env("SIPI_USE_LOCAL_DEV", "false")
        .with_env("ALLOW_ERASE_PROJECTS", "true")
        .with_env("DB_JDBC_URL", "jdbc:sqlite:/opt/db/ingest.sqlite")
        .with_volume_mapping(SIPI_IMAGES, "/opt/images", "rw")
        .with_volume_mapping(TMP_INGEST, "/opt/temp", "rw")
        .with_volume_mapping(INGEST_DB, "/opt/db", "rw")
    )
    ingest.start()
    wait_for_logs(ingest, "Started dsp-ingest")
    print("Ingest is ready")
    return ingest


def _get_api_container(network: Network, version: str) -> DockerContainer:
    api = (
        DockerContainer(f"daschswiss/knora-api:{version}")
        .with_name("api")
        .with_network(network)
        .with_env("KNORA_WEBAPI_DSP_INGEST_BASE_URL", "http://ingest:3340")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_HOST", "db")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_REPOSITORY_NAME", "knora-test")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_USERNAME", "admin")
        .with_env("KNORA_WEBAPI_TRIPLESTORE_FUSEKI_PASSWORD", "test")
        .with_env("ALLOW_ERASE_PROJECTS", "true")
        .with_bind_ports(3333, 3333)
    )
    api.start()
    wait_for_logs(api, "AppState set to Running")
    wait_for_logs(api, "Starting api on")
    print("API is ready")
    return api


def _print_containers_are_ready(containers: Containers) -> None:
    print("Containers are ready")
    print(f"  {containers.fuseki._name}: {containers.fuseki.ports}")
    print(f"  {containers.sipi._name}: {containers.sipi.ports}")
    print(f"  {containers.ingest._name}: {containers.ingest.ports}")
    print(f"  {containers.api._name}: {containers.api.ports}")


def _stop_all_containers(containers: Containers) -> None:
    containers.fuseki.stop()
    containers.sipi.stop()
    containers.ingest.stop()
    containers.api.stop()
    print("All containers have been stopped")



import copy
import dataclasses
import datetime
import difflib
import json
import os
import uuid
import warnings
from pathlib import Path
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Union

import regex
from lxml import etree
from lxml.builder import E
from namedentities.core import numeric_entities  # type: ignore[import-untyped]
from regex import Match

from dsp_tools.commands.excel2xml.propertyelement import PropertyElement
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.models.datetimestamp import DateTimeStamp
from dsp_tools.models.exceptions import BaseError
from dsp_tools.utils.date_util import is_full_date
from dsp_tools.utils.shared import check_notna
from dsp_tools.utils.shared import simplify_name
from dsp_tools.utils.uri_util import is_iiif_uri
from dsp_tools.utils.uri_util import is_uri
from dsp_tools.utils.xml_validation import validate_xml_file

# ruff: noqa: E501, UP031 (line-too-long, use f-string over percent formatting)


xml_namespace_map = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}


def make_xsd_id_compatible(string: str) -> str:
    """
    An xsd:ID may not contain all types of special characters.
    This function replaces illegal characters with "_" and appends a random UUID.
    The UUID will be different each time the function is called.

    The string must contain at least one Unicode letter (matching the regex ``\\p{L}``),
    underscore, !, ?, or number, but must not be `None`, `<NA>`, `N/A`, or `-`.
    Otherwise, a BaseError will be raised.

    Args:
        string: input string

    Raises:
        BaseError: if the input cannot be transformed to an xsd:ID

    Returns:
        an xsd ID based on the input string, with a UUID attached.
    """
    res = make_xsd_id_compatible_without_uuid(string)
    _uuid = uuid.uuid4()
    res = f"{res}_{_uuid}"
    return res


def make_xsd_id_compatible_without_uuid(string: str) -> str:
    """
    An xsd:ID may not contain all types of special characters.
    This function replaces illegal characters with "_".

    The string must contain at least one Unicode letter (matching the regex ``\\p{L}``),
    underscore, !, ?, or number, but must not be `None`, `<NA>`, `N/A`, or `-`.
    Otherwise, a BaseError will be raised.

    Args:
        string: input string

    Raises:
        BaseError: if the input cannot be transformed to an xsd:ID

    Returns:
        an xsd ID compatible string based on the input string
    """
    if not isinstance(string, str) or not check_notna(string):
        raise BaseError(f"The input '{string}' cannot be transformed to an xsd:ID")
    # if start of string is neither letter nor underscore, add an underscore
    res = regex.sub(r"^(?=[^A-Za-z_])", "_", string)
    # replace all illegal characters by underscore
    res = regex.sub(r"[^\w_\-.]", "_", res, flags=regex.ASCII)
    return res


def _find_french_bc_date(
    string: str,
    lookbehind: str,
    lookahead: str,
) -> Optional[str]:
    french_bc_regex = r"av(?:\. |\.| )J\.?-?C\.?"
    if not regex.search(french_bc_regex, string):
        return None

    year_regex = r"\d{1,5}"
    sep_regex = r" ?- ?"

    year_range_regex = rf"{lookbehind}({year_regex}){sep_regex}({year_regex}) {french_bc_regex}{lookahead}"
    year_range = regex.search(year_range_regex, string)
    if year_range:
        start_year = int(year_range.group(1))
        end_year = int(year_range.group(2))
        if end_year > start_year:
            return None
        return f"GREGORIAN:BC:{start_year}:BC:{end_year}"

    single_year_regex = rf"{lookbehind}({year_regex}) {french_bc_regex}{lookahead}"
    single_year = regex.search(single_year_regex, string)
    if single_year:
        start_year = int(single_year.group(1))
        return f"GREGORIAN:BC:{start_year}:BC:{start_year}"

    return None


def find_date_in_string(string: str) -> Optional[str]:
    """
    Checks if a string contains a date value (single date, or date range),
    and returns the first found date as DSP-formatted string.
    Returns None if no date was found.

    Notes:
        - All dates are interpreted in the Christian era and the Gregorian calendar.
        - BC dates are only supported in French notation (e.g. 1000-900 av. J.-C.).
        - The years 0000-2999 are supported, in 3/4-digit form.
        - Dates written with slashes are always interpreted in a European manner: 5/11/2021 is the 5th of November.
        - In the European notation, 2-digit years are expanded to 4 digits, with the current year as watershed:
            - 30.4.24 -> 30.04.2024
            - 30.4.25 -> 30.04.1925

    Currently supported date formats:
        - 0476-09-04 -> GREGORIAN:CE:0476-09-04:CE:0476-09-04
        - 0476_09_04 -> GREGORIAN:CE:0476-09-04:CE:0476-09-04
        - 30.4.2021 -> GREGORIAN:CE:2021-04-30:CE:2021-04-30
        - 30.4.21 -> GREGORIAN:CE:2021-04-30:CE:2021-04-30
        - 5/11/2021 -> GREGORIAN:CE:2021-11-05:CE:2021-11-05
        - Jan 26, 1993 -> GREGORIAN:CE:1993-01-26:CE:1993-01-26
        - 28.2.-1.12.1515 -> GREGORIAN:CE:1515-02-28:CE:1515-12-01
        - 25.-26.2.0800 -> GREGORIAN:CE:0800-02-25:CE:0800-02-26
        - 1.9.2022-3.1.2024 -> GREGORIAN:CE:2022-09-01:CE:2024-01-03
        - 1848 -> GREGORIAN:CE:1848:CE:1848
        - 1849/1850 -> GREGORIAN:CE:1849:CE:1850
        - 1849/50 -> GREGORIAN:CE:1849:CE:1850
        - 1845-50 -> GREGORIAN:CE:1845:CE:1850
        - 840-50 -> GREGORIAN:CE:840:CE:850
        - 840-1 -> GREGORIAN:CE:840:CE:841
        - 1000-900 av. J.-C. -> GREGORIAN:BC:1000:BC:900
        - 45 av. J.-C. -> GREGORIAN:BC:45:BC:45

    Args:
        string: string to check

    Returns:
        DSP-formatted date string, or None

    Examples:
        >>> if find_date_in_string(row["Epoch"]):
        >>>     resource.append(make_date_prop(":hasDate", find_date_in_string(row["Epoch"]))

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date-prop
    """

    # sanitize input, just in case that the method was called on an empty or N/A cell
    if not check_notna(string):
        return None
    try:
        return _find_date_in_string_throwing(string)
    except ValueError:
        return None


_months_dict = {
    "January": 1,
    "Jan": 1,
    "February": 2,
    "Feb": 2,
    "March": 3,
    "Mar": 3,
    "April": 4,
    "Apr": 4,
    "May": 5,
    "June": 6,
    "Jun": 6,
    "July": 7,
    "Jul": 7,
    "August": 8,
    "Aug": 8,
    "September": 9,
    "Sept": 9,
    "October": 10,
    "Oct": 10,
    "November": 11,
    "Nov": 11,
    "December": 12,
    "Dec": 12,
}


def _find_date_in_string_throwing(string: str) -> str | None:
    """
    This function is the same as find_date_in_string(), but may raise a ValueError instead of returning None.
    """
    year_regex = r"([0-2]?[0-9][0-9][0-9])"
    year_regex_2_or_4_digits = r"((?:[0-2]?[0-9])?[0-9][0-9])"
    month_regex = r"([0-1]?[0-9])"
    day_regex = r"([0-3]?[0-9])"
    sep_regex = r"[\./]"
    lookbehind = r"(?<![0-9A-Za-z])"
    lookahead = r"(?![0-9A-Za-z])"

    if french_bc_date := _find_french_bc_date(string=string, lookbehind=lookbehind, lookahead=lookahead):
        return french_bc_date

    # template: 2021-01-01 | 2015_01_02
    iso_date = regex.search(rf"{lookbehind}{year_regex}[_-]([0-1][0-9])[_-]([0-3][0-9]){lookahead}", string)

    # template: 6.-8.3.1948 | 6/2/1947 - 24.03.1948
    eur_date_range_regex = (
        rf"{lookbehind}"
        rf"{day_regex}{sep_regex}(?:{month_regex}{sep_regex}{year_regex_2_or_4_digits}?)? ?(?:-|:|to) ?"
        rf"{day_regex}{sep_regex}{month_regex}{sep_regex}{year_regex_2_or_4_digits}"
        rf"{lookahead}"
    )
    eur_date_range = regex.search(eur_date_range_regex, string)

    # template: 1.4.2021 | 5/11/2021
    eur_date_regex = rf"{lookbehind}{day_regex}{sep_regex}{month_regex}{sep_regex}{year_regex_2_or_4_digits}{lookahead}"
    eur_date = regex.search(
        eur_date_regex,
        string,
    )

    # template: March 9, 1908 | March5,1908 | May 11, 1906
    all_months = "|".join(_months_dict)
    monthname_date_regex = rf"{lookbehind}({all_months}) ?{day_regex}, ?{year_regex}{lookahead}"
    monthname_date = regex.search(monthname_date_regex, string)

    # template: 1849/50 | 1849-50 | 1849/1850
    year_range = regex.search(lookbehind + year_regex + r"[/-](\d{1,4})" + lookahead, string)

    # template: 1907
    year_only = regex.search(rf"{lookbehind}{year_regex}{lookahead}", string)

    res: str | None = None
    if iso_date:
        res = _from_iso_date(iso_date)
    elif eur_date_range:
        res = _from_eur_date_range(eur_date_range)
    elif eur_date:
        res = _from_eur_date(eur_date)
    elif monthname_date:
        res = _from_monthname_date(monthname_date)
    elif year_range:
        res = _from_year_range(year_range)
    elif year_only:
        year = int(year_only.group(0))
        res = f"GREGORIAN:CE:{year}:CE:{year}"
    return res


def _from_iso_date(iso_date: Match[str]) -> str:
    year = int(iso_date.group(1))
    month = int(iso_date.group(2))
    day = int(iso_date.group(3))
    date = datetime.date(year, month, day)
    return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"


def _expand_2_digit_year(year: int) -> int:
    current_year = datetime.date.today().year - 2000
    if year <= current_year:
        return year + 2000
    elif year <= 99:
        return year + 1900
    else:
        return year


def _from_eur_date_range(eur_date_range: Match[str]) -> str:
    startday = int(eur_date_range.group(1))
    startmonth = int(eur_date_range.group(2)) if eur_date_range.group(2) else int(eur_date_range.group(5))
    startyear = int(eur_date_range.group(3)) if eur_date_range.group(3) else int(eur_date_range.group(6))
    startyear = _expand_2_digit_year(startyear)
    endday = int(eur_date_range.group(4))
    endmonth = int(eur_date_range.group(5))
    endyear = int(eur_date_range.group(6))
    endyear = _expand_2_digit_year(endyear)
    startdate = datetime.date(startyear, startmonth, startday)
    enddate = datetime.date(endyear, endmonth, endday)
    if enddate < startdate:
        raise ValueError
    return f"GREGORIAN:CE:{startdate.isoformat()}:CE:{enddate.isoformat()}"


def _from_eur_date(eur_date: Match[str]) -> str:
    startday = int(eur_date.group(1))
    startmonth = int(eur_date.group(2))
    startyear = int(eur_date.group(3))
    startyear = _expand_2_digit_year(startyear)
    date = datetime.date(startyear, startmonth, startday)
    return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"


def _from_monthname_date(monthname_date: Match[str]) -> str:
    day = int(monthname_date.group(2))
    month = _months_dict[monthname_date.group(1)]
    year = int(monthname_date.group(3))
    date = datetime.date(year, month, day)
    return f"GREGORIAN:CE:{date.isoformat()}:CE:{date.isoformat()}"


def _from_year_range(year_range: Match[str]) -> str:
    startyear = int(year_range.group(1))
    endyear = int(year_range.group(2))
    if endyear // 10 == 0:
        # endyear is only 1-digit: add the first 2-3 digits of startyear
        endyear = startyear // 10 * 10 + endyear
    elif endyear // 100 == 0:
        # endyear is only 2-digit: add the first 1-2 digits of startyear
        endyear = startyear // 100 * 100 + endyear
    if endyear <= startyear:
        raise ValueError
    return f"GREGORIAN:CE:{startyear}:CE:{endyear}"


def prepare_value(
    value: Union[PropertyElement, str, int, float, bool, Iterable[Union[PropertyElement, str, int, float, bool]]],
) -> list[PropertyElement]:
    """
    This method transforms the parameter "value" from a make_*_prop() method into a list of PropertyElements. "value" is
    passed on to this method as it was received.

    Args:
        value: "value" as received from the caller

    Returns:
        a list of PropertyElements
    """
    # make sure that "value" is list-like
    if not isinstance(value, Iterable) or isinstance(value, str):
        value = [value]

    # make a PropertyElement out of its elements, if necessary.
    return [x if isinstance(x, PropertyElement) else PropertyElement(x) for x in value]


def make_root(
    shortcode: str,
    default_ontology: str,
) -> etree._Element:
    """
    Start building your XML document by creating the root element `<knora>`.

    Args:
        shortcode: The shortcode of this project as defined in the JSON project file
        default_ontology: one of the ontologies of the JSON project file

    Returns:
        The root element `<knora>`.

    Examples:
        >>> root = excel2xml.make_root(shortcode=shortcode, default_ontology=default_ontology)
        >>> root = excel2xml.append_permissions(root)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#the-root-element-knora
    """
    schema_url = "https://raw.githubusercontent.com/dasch-swiss/dsp-tools/main/src/dsp_tools/resources/schema/data.xsd"
    schema_location_key = str(etree.QName("http://www.w3.org/2001/XMLSchema-instance", "schemaLocation"))
    schema_location_value = f"https://dasch.swiss/schema {schema_url}"
    return etree.Element(
        "{%s}knora" % xml_namespace_map[None],
        attrib={
            schema_location_key: schema_location_value,
            "shortcode": shortcode,
            "default-ontology": default_ontology,
        },
        nsmap=xml_namespace_map,
    )


def append_permissions(root_element: etree._Element) -> etree._Element:
    """
    After having created a root element, call this function to append the standard permission definitions "open",
    "restricted-view", and "restricted" to it. These definitions are a good basis to
    start with, but remember that they can be adapted, and that other permissions can be defined instead of these.

    Args:
        root_element: The XML root element `<knora>` created by make_root()

    Returns:
        The root element with the permission definition blocks appended

    Examples:
        >>> root = excel2xml.make_root(shortcode=shortcode, default_ontology=default_ontology)
        >>> root = excel2xml.append_permissions(root)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#describing-permissions-with-permissions-elements
    """

    PERMISSIONS = E.permissions
    ALLOW = E.allow
    # lxml.builder.E is a more sophisticated element factory than etree.Element.
    # E.tag is equivalent to E("tag") and results in <tag>

    _open = PERMISSIONS(id="open")
    _open.append(ALLOW("V", group="UnknownUser"))
    _open.append(ALLOW("V", group="KnownUser"))
    _open.append(ALLOW("D", group="ProjectMember"))
    _open.append(ALLOW("CR", group="ProjectAdmin"))
    root_element.append(_open)

    _restricted_view = PERMISSIONS(id="restricted-view")
    _restricted_view.append(ALLOW("RV", group="UnknownUser"))
    _restricted_view.append(ALLOW("RV", group="KnownUser"))
    _restricted_view.append(ALLOW("D", group="ProjectMember"))
    _restricted_view.append(ALLOW("CR", group="ProjectAdmin"))
    root_element.append(_restricted_view)

    _restricted = PERMISSIONS(id="restricted")
    _restricted.append(ALLOW("D", group="ProjectMember"))
    _restricted.append(ALLOW("CR", group="ProjectAdmin"))
    root_element.append(_restricted)

    return root_element


def make_resource(  # noqa: D417 (undocumented-param)
    label: str,
    restype: str,
    id: str,
    permissions: str = "open",
    ark: Optional[str] = None,
    iri: Optional[str] = None,
    creation_date: Optional[str] = None,
) -> etree._Element:
    """
    Creates an empty resource element, with the attributes as specified by the arguments.

    Args:
        The arguments correspond to the attributes of the `<resource>` element.

    Returns:
        The resource element, without any children, but with the attributes
        ``<resource label=label restype=restype id=id permissions=permissions ark=ark iri=iri></resource>``

    Raises:
        Warning: if both an ARK and an IRI are provided
        BaseError: if the creation date is invalid

    Examples:
        >>> resource = excel2xml.make_resource(...)
        >>> resource.append(excel2xml.make_text_prop(...))
        >>> root.append(resource)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#describing-resources-with-the-resource-element
    """
    if not check_notna(label):
        msg = f"Your resource's label looks suspicious (resource with id '{id}' and label '{label}')"
        warnings.warn(DspToolsUserWarning(msg))
    if not check_notna(id):
        msg = f"Your resource's id looks suspicious (resource with id '{id}' and label '{label}'"
        warnings.warn(DspToolsUserWarning(msg))
    kwargs = {"label": label, "restype": restype, "id": id, "permissions": permissions, "nsmap": xml_namespace_map}
    if ark:
        kwargs["ark"] = ark
    if iri:
        kwargs["iri"] = iri
    if ark and iri:
        msg = f"Both ARK and IRI were provided for resource '{label}' ({id}). The ARK will override the IRI."
        warnings.warn(DspToolsUserWarning(msg))
    if creation_date:
        try:
            DateTimeStamp(creation_date)
        except BaseError:
            raise BaseError(
                f"The resource '{label}' (ID: {id}) has an invalid creation date '{creation_date}'. "
                f"Did you perhaps forget the timezone?"
            ) from None
        kwargs["creation_date"] = creation_date

    return etree.Element("{%s}resource" % xml_namespace_map[None], **kwargs)  # type: ignore[arg-type]


def make_bitstream_prop(
    path: Union[str, os.PathLike[Any]],
    permissions: str = "open",
    check: bool = False,
    calling_resource: str = "",
) -> etree._Element:
    """
    Creates a bitstream element that points to "path".

    Args:
        path: path to a valid file that will be uploaded
        permissions: permissions string
        check: if True, issue a warning if the path doesn't point to an existing file
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        if the path doesn't point to an existing file (only if check=True)

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> resource = excel2xml.make_resource(...)
        >>> resource.append(excel2xml.make_bitstream_prop("data/images/tree.jpg"))
        >>> root.append(resource)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#bitstream
    """

    if check and not Path(path).is_file():
        msg = (
            f"Failed validation in bitstream tag of resource '{calling_resource}': "
            f"The following path doesn't point to a file: {path}"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop_ = etree.Element(
        "{%s}bitstream" % xml_namespace_map[None],
        permissions=permissions,
        nsmap=xml_namespace_map,
    )
    prop_.text = str(path)
    return prop_


def make_iiif_uri_prop(
    iiif_uri: str,
    permissions: str = "open",
    calling_resource: str = "",
) -> etree._Element:
    """
    Creates a iiif-uri element that points to "path".

    Args:
        iiif_uri: URI to a IIIF image
        permissions: permissions string
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the iiif_uri doesn't conform to the IIIF URI specifications

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> resource = excel2xml.make_resource(...)
        >>> resource.append(excel2xml.make_iiif_uri_prop("https://example.org/image-service/abcd1234/full/max/0/default.jpg"))
        >>> root.append(resource)

    """

    if not is_iiif_uri(iiif_uri):
        msg = (
            f"Failed validation in iiif-uri tag of resource '{calling_resource}': "
            f"The URI: '{iiif_uri}' does not conform to the specifications."
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop_ = etree.Element(
        "{%s}iiif-uri" % xml_namespace_map[None],
        permissions=permissions,
        nsmap=xml_namespace_map,
    )
    prop_.text = iiif_uri
    return prop_


def _format_bool(
    unformatted: Union[bool, str, int, float],
    name: str,
    calling_resource: str,
) -> str:
    """
    This method takes an unformatted boolean-like value, and transforms it into the string values "true" or "false".

    Args:
        unformatted: boolean-like value
        name: property name, for better error messages
        calling_resource: resource name, for better error messages

    Raises:
        BaseError: if the input cannot be transformed into "true"/"false"

    Returns:
        "true" if the input is in (True, "true", "1", 1, "yes"); "false" if input is in (False, "false", "0", 0, "no")
    """
    if isinstance(unformatted, str):
        unformatted = unformatted.lower()
    if unformatted in (False, "false", "0", 0, 0.0, "no"):
        return "false"
    elif unformatted in (True, "true", "1", 1, 1.0, "yes"):
        return "true"
    else:
        raise BaseError(
            f"Failed validation in resource '{calling_resource}', property '{name}': "
            f"'{unformatted}' is not a valid boolean."
        )


def make_boolean_prop(
    name: str,
    value: Union[PropertyElement, str, int, bool],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<boolean-prop>` from a boolean value. The value can be provided directly or inside a PropertyElement. The
    following formats are supported:
     - true: (True, "true", "True", "1", 1, "yes", "Yes")
     - false: (False, "false", "False", "0", 0, "no", "No")

    Unless provided as PropertyElement, the permissions of the value default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: a boolean value as str/bool/int, or as str/bool/int inside a PropertyElement
        calling_resource: the name of the parent resource (for better error messages)

    Raises:
        BaseError: if the value is not a valid boolean

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_boolean_prop(":testproperty", "no")
                <boolean-prop name=":testproperty">
                    <boolean permissions="open">false</boolean>
                </boolean-prop>
        >>> excel2xml.make_boolean_prop(":testproperty", excel2xml.PropertyElement("1", permissions="restricted", comment="example"))
                <boolean-prop name=":testproperty">
                    <boolean permissions="restricted" comment="example">true</boolean>
                </boolean-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#boolean-prop
    """

    # validate input
    if isinstance(value, PropertyElement):
        value_new = dataclasses.replace(value, value=_format_bool(value.value, name, calling_resource))
    elif isinstance(value, (str, bool, int)):
        value_new = PropertyElement(_format_bool(value, name, calling_resource))
    else:
        raise BaseError(
            f"Failed validation in resource '{calling_resource}', property '{name}': '{value}' is not a valid boolean."
        )

    # make xml structure of the value
    prop_ = etree.Element(
        "{%s}boolean-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    kwargs = {"permissions": value_new.permissions}
    if value_new.comment and check_notna(value_new.comment):
        kwargs["comment"] = value_new.comment
    value_ = etree.Element(
        "{%s}boolean" % xml_namespace_map[None],
        **kwargs,  # type: ignore[arg-type]
        nsmap=xml_namespace_map,
    )
    value_.text = str(value_new.value)
    prop_.append(value_)

    return prop_


def make_color_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<color-prop>` from one or more colors. The color(s) can be provided as string or as PropertyElement with a
    string inside. If provided as string, the permissions default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more DSP color(s), as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid color

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_color_prop(":testproperty", "#00ff66")
                <color-prop name=":testproperty">
                    <color permissions="open">#00ff66</color>
                </color-prop>
        >>> excel2xml.make_color_prop(":testproperty", excel2xml.PropertyElement("#00ff66", permissions="restricted", comment="example"))
                <color-prop name=":testproperty">
                    <color permissions="restricted" comment="example">#00ff66</color>
                </color-prop>
        >>> excel2xml.make_color_prop(":testproperty", ["#00ff66", "#000000"])
                <color-prop name=":testproperty">
                    <color permissions="open">#00ff66</color>
                    <color permissions="open">#000000</color>
                </color-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#color-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not regex.search(r"^#[0-9a-f]{6}$", str(val.value).strip(), flags=regex.IGNORECASE):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid color."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}color-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}color" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value).strip()
        prop_.append(value_)

    return prop_


def make_date_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<date-prop>` from one or more dates/date ranges. The date(s) can be provided as string or as PropertyElement
    with a string inside. If provided as string, the permissions default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more DSP dates, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid DSP date

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_date_prop(":testproperty", "GREGORIAN:CE:2014-01-31")
                <date-prop name=":testproperty">
                    <date permissions="open">GREGORIAN:CE:2014-01-31</date>
                </date-prop>
        >>> excel2xml.make_date_prop(":testproperty", excel2xml.PropertyElement("GREGORIAN:CE:2014-01-31", permissions="restricted", comment="example"))
                <date-prop name=":testproperty">
                    <date permissions="restricted" comment="example">
                        GREGORIAN:CE:2014-01-31
                    </date>
                </date-prop>
        >>> excel2xml.make_date_prop(":testproperty", ["GREGORIAN:CE:1930-09-02:CE:1930-09-03", "GREGORIAN:CE:1930-09-02:CE:1930-09-03"])
                <date-prop name=":testproperty">
                    <date permissions="open">
                        GREGORIAN:CE:1930-09-02:CE:1930-09-03
                    </date>
                    <date permissions="open">
                        GREGORIAN:CE:1930-09-02:CE:1930-09-03
                    </date>
                </date-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not is_full_date(str(val.value).strip()):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid DSP date."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}date-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}date" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value).strip()
        prop_.append(value_)

    return prop_


def make_decimal_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<decimal-prop>` from one or more decimal numbers. The decimal(s) can be provided as string, float, or as
    PropertyElement with a string/float inside. If provided as string/float, the permissions default to
    "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more decimal numbers, as string/float/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid decimal number

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_decimal_prop(":testproperty", "3.14159")
                <decimal-prop name=":testproperty">
                    <decimal permissions="open">3.14159</decimal>
                </decimal-prop>
        >>> excel2xml.make_decimal_prop(":testproperty", excel2xml.PropertyElement("3.14159", permissions="restricted", comment="example"))
                <decimal-prop name=":testproperty">
                    <decimal permissions="restricted" comment="example">3.14159</decimal>
                </decimal-prop>
        >>> excel2xml.make_decimal_prop(":testproperty", ["3.14159", "2.718"])
                <decimal-prop name=":testproperty">
                    <decimal permissions="open">3.14159</decimal>
                    <decimal permissions="open">2.718</decimal>
                </decimal-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#decimal-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        try:
            float(val.value)
        except ValueError:
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid decimal number."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}decimal-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}decimal" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_geometry_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<geometry-prop>` from one or more areas of an image. The area(s) can be provided as JSON-string or as
    PropertyElement with the JSON-string inside. If provided as string, the permissions default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more JSON geometry objects, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid JSON geometry object

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_geometry_prop(":testproperty", json_string)
                <geometry-prop name=":testproperty">
                    <geometry permissions="open">{JSON}</geometry>
                </geometry-prop>
        >>> excel2xml.make_geometry_prop(":testproperty", excel2xml.PropertyElement(json_string, permissions="restricted", comment="example"))
                <geometry-prop name=":testproperty">
                    <geometry permissions="restricted" comment="example">{JSON}</geometry>
                </geometry-prop>
        >>> excel2xml.make_geometry_prop(":testproperty", [json_string1, json_string2])
                <geometry-prop name=":testproperty">
                    <geometry permissions="open">{JSON}</geometry>
                    <geometry permissions="open">{JSON}</geometry>
                </geometry-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geometry-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        try:
            value_as_dict = json.loads(str(val.value))
            if value_as_dict["type"] not in ["rectangle", "circle", "polygon"]:
                msg = (
                    f"Failed validation in resource '{calling_resource}', property '{name}': "
                    f"The 'type' of the JSON geometry object must be 'rectangle', 'circle', or 'polygon'."
                )
                warnings.warn(DspToolsUserWarning(msg))
            if not isinstance(value_as_dict["points"], list):
                msg = (
                    f"Failed validation in resource '{calling_resource}', property '{name}': "
                    f"The 'points'of the JSON geometry object must be a list of points."
                )
                warnings.warn(DspToolsUserWarning(msg))
        except (json.JSONDecodeError, TypeError, IndexError, KeyError, AssertionError):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid JSON geometry object."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}geometry-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}geometry" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)
    return prop_


def make_geoname_prop(
    name: str,
    value: Union[PropertyElement, str, int, Iterable[Union[PropertyElement, str, int]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<geoname-prop>` from one or more geonames.org IDs. The ID(s) can be provided as string, integer, or as
    PropertyElement with a string/integer inside. If provided as string/integer, the permissions default to
    "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more geonames.org IDs, as str/int/PropertyElement, or as iterable of str/int/PropertyElement
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid geonames.org identifier

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_geoname_prop(":testproperty", "2761369")
                <geoname-prop name=":testproperty">
                    <geoname permissions="open">2761369</geoname>
                </geoname-prop>
        >>> excel2xml.make_geoname_prop(":testproperty", excel2xml.PropertyElement("2761369", permissions="restricted", comment="example"))
                <geoname-prop name=":testproperty">
                    <geoname permissions="restricted" comment="example">2761369</geoname>
                </geoname-prop>
        >>> excel2xml.make_geoname_prop(":testproperty", ["2761369", "1010101"])
                <geoname-prop name=":testproperty">
                    <geoname permissions="open">2761369</geoname>
                    <geoname permissions="open">1010101</geoname>
                </geoname-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geoname-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not regex.search(r"^[0-9]+$", str(val.value)):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a geonames.org identifier."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}geoname-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}geoname" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_integer_prop(
    name: str,
    value: Union[PropertyElement, str, int, Iterable[Union[PropertyElement, str, int]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<integer-prop>` from one or more integers. The integers can be provided as string, integer, or as
    PropertyElement with a string/integer inside. If provided as string/integer, the permissions default to
    "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more integers, as string/int/PropertyElement, or as iterable of strings/ints/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the value is not a valid integer

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_integer_prop(":testproperty", "2761369")
                <integer-prop name=":testproperty">
                    <integer permissions="open">2761369</integer>
                </integer-prop>
        >>> excel2xml.make_integer_prop(":testproperty", excel2xml.PropertyElement("2761369", permissions="restricted", comment="example"))
                <integer-prop name=":testproperty">
                    <integer permissions="restricted" comment="example">2761369</integer>
                </integer-prop>
        >>> excel2xml.make_integer_prop(":testproperty", ["2761369", "1010101"])
                <integer-prop name=":testproperty">
                    <integer permissions="open">2761369</integer>
                    <integer permissions="open">1010101</integer>
                </integer-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#integer-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        try:
            int(val.value)
        except ValueError:
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid integer."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}integer-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}integer" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_list_prop(
    list_name: str,
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<list-prop>` from one or more list nodes. The name(s) of the list node(s) can be provided as string or as
    PropertyElement with a string inside. If provided as string, the permissions default to "open".

    Args:
        list_name: the name of the list as defined in the onto
        name: the name of this property as defined in the onto
        value: one or more node names, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the name of one of the list nodes is not a valid string

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_list_prop("mylist", ":testproperty", "first_node")
                <list-prop list="mylist" name=":testproperty">
                    <list permissions="open">first_node</list>
                </list-prop>
        >>> excel2xml.make_list_prop("mylist", ":testproperty", excel2xml.PropertyElement("first_node", permissions="restricted", comment="example"))
                <list-prop list="mylist" name=":testproperty">
                    <list permissions="restricted" comment="example">first_node</list>
                </list-prop>
        >>> excel2xml.make_list_prop("mylist", ":testproperty", ["first_node", "second_node"])
                <list-prop list="mylist" name=":testproperty">
                    <list permissions="open">first_node</list>
                    <list permissions="open">second_node</list>
                </list-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#list-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not isinstance(val.value, str) or not check_notna(val.value):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid name of a list node."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}list-prop" % xml_namespace_map[None],
        list=list_name,
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}list" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_resptr_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<resptr-prop>` from one or more IDs of other resources. The ID(s) can be provided as string or as
    PropertyElement with a string inside. If provided as string, the permissions default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more resource identifiers, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If the ID of one of the target resources is not a valid string

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_resptr_prop(":testproperty", "resource_1")
                <resptr-prop name=":testproperty">
                    <resptr permissions="open">resource_1</resptr>
                </resptr-prop>
        >>> excel2xml.make_resptr_prop(":testproperty", excel2xml.PropertyElement("resource_1", permissions="restricted", comment="example"))
                <resptr-prop name=":testproperty">
                    <resptr permissions="restricted" comment="example">resource_1</resptr>
                </resptr-prop>
        >>> excel2xml.make_resptr_prop(":testproperty", ["resource_1", "resource_2"])
                <resptr-prop name=":testproperty">
                    <resptr permissions="open">resource_1</resptr>
                    <resptr permissions="open">resource_2</resptr>
                </resptr-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#resptr-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not isinstance(val.value, str) or not check_notna(val.value):
            msg = (
                f"Validation Error in resource '{calling_resource}', property '{name}': "
                f"The following doesn't seem to be a valid ID of a target resource: '{val.value}'"
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}resptr-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}resptr" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_text_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<text-prop>` from one or more strings. The string(s) can be provided as string or as PropertyElement with a
    string inside. If provided as string, the encoding defaults to utf8, and the permissions to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more strings, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Raises:
        BaseError: if the XML tags in a richtext property (encoding=xml) are not well-formed
        Warning: if one of the values doesn't look like a reasonable string
            (e.g. `<NA>` is a valid string, but probably not intended)

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_text_prop(":testproperty", "first text")
                <text-prop name=":testproperty">
                    <text encoding="utf8" permissions="open">first text</text>
                </text-prop>
        >>> excel2xml.make_text_prop(":testproperty", excel2xml.PropertyElement("first text", permissions="restricted", encoding="xml"))
                <text-prop name=":testproperty">
                    <text encoding="xml" permissions="restricted">first text</text>
                </text-prop>
        >>> excel2xml.make_text_prop(":testproperty", ["first text", "second text"])
                <text-prop name=":testproperty">
                    <text encoding="utf8" permissions="open">first text</text>
                    <text encoding="utf8" permissions="open">second text</text>
                </text-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not isinstance(val.value, str) or not check_notna(val.value):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is probably not a usable string."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}text-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if check_notna(val.comment):
            kwargs["comment"] = val.comment
        kwargs["encoding"] = val.encoding if check_notna(val.encoding) else "utf8"
        value_ = etree.Element(
            "{%s}text" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        if kwargs["encoding"] == "utf8":
            # write the text into the tag, without validation
            value_.text = str(val.value)
        else:
            try:
                value_ = _add_richtext_to_etree_element(str(val.value), value_)
            except BaseError as err:
                if calling_resource:
                    err.message += f"The error occurred in resource {calling_resource}, property {name}"
                raise err from None
        prop_.append(value_)

    return prop_


def _add_richtext_to_etree_element(richtext: str, element: etree._Element) -> etree._Element:
    new_element = copy.deepcopy(element)
    escaped_text = _escape_reserved_chars(richtext)
    # transform named entities (=character references) to numeric entities, e.g. &nbsp; -> &#160;
    num_ent = numeric_entities(escaped_text)
    pseudo_xml = f"<ignore-this>{num_ent}</ignore-this>"
    try:
        parsed = etree.fromstring(pseudo_xml)
    except etree.XMLSyntaxError as err:
        msg = (
            "The XML tags contained in a richtext property (encoding=xml) must be well-formed. "
            "The special characters <, > and & are only allowed to construct a tag. "
        )
        msg += f"\nOriginal error message: {err.msg}"
        msg += f"\nEventual line/column numbers are relative to this text: {pseudo_xml}"
        raise BaseError(msg) from None
    new_element.text = parsed.text  # everything before the first child tag
    new_element.extend(list(parsed))  # all (nested) children of the pseudo-xml
    return new_element


def _escape_reserved_chars(text: str) -> str:
    """
    From richtext strings (encoding="xml"), escape the reserved characters <, > and &,
    but only if they are not part of a standard standoff tag or escape sequence.
    The standard standoff tags allowed by DSP-API are documented here:
    https://docs.dasch.swiss/2023.12.01/DSP-API/03-endpoints/api-v2/text/standard-standoff/

    Args:
        text: the richtext string to be escaped

    Returns:
        the escaped richtext string
    """
    allowed_tags = [
        "a( [^>]+)?",  # <a> is the only tag that can have attributes
        "p",
        "em",
        "strong",
        "u",
        "sub",
        "sup",
        "strike",
        "h1",
        "ol",
        "ul",
        "li",
        "tbody",
        "table",
        "tr",
        "td",
        "br",
        "hr",
        "pre",
        "cite",
        "blockquote",
        "code",
    ]
    allowed_tags_regex = "|".join(allowed_tags)
    lookahead = rf"(?!/?({allowed_tags_regex})/?>)"
    illegal_lt = rf"<{lookahead}"
    lookbehind = rf"(?<!</?({allowed_tags_regex})/?)"
    illegal_gt = rf"{lookbehind}>"
    illegal_amp = r"&(?![#a-zA-Z0-9]+;)"
    text = regex.sub(illegal_lt, "&lt;", text)
    text = regex.sub(illegal_gt, "&gt;", text)
    text = regex.sub(illegal_amp, "&amp;", text)
    return text


def make_time_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<time-prop>` from one or more datetime values of the form "2009-10-10T12:00:00-05:00". The time(s) can be
    provided as string or as PropertyElement with a string inside. If provided as string, the permissions default to
    "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more DSP times, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If one of the values is not a valid DSP time string

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_time_prop(":testproperty", "2009-10-10T12:00:00-05:00")
                <time-prop name=":testproperty">
                    <time permissions="open">
                        2009-10-10T12:00:00-05:00
                    </time>
                </time-prop>
        >>> excel2xml.make_time_prop(":testproperty", excel2xml.PropertyElement("2009-10-10T12:00:00-05:00", permissions="restricted", comment="example"))
                <time-prop name=":testproperty">
                    <time permissions="restricted" comment="example">
                        2009-10-10T12:00:00-05:00
                    </time>
                </time-prop>
        >>> excel2xml.make_time_prop(":testproperty", ["2009-10-10T12:00:00-05:00", "1901-01-01T01:00:00-00:00"])
                <time-prop name=":testproperty">
                    <time permissions="open">
                        2009-10-10T12:00:00-05:00
                    </time>
                    <time permissions="open">
                        1901-01-01T01:00:00-00:002
                    </time>
                </time-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#time-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    validation_regex = r"^\d{4}-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:[0-5]\d(.\d{1,12})?(Z|[+-][0-1]\d:[0-5]\d)$"
    for val in values:
        if not regex.search(validation_regex, str(val.value)):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid DSP time."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}time-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}time" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_uri_prop(
    name: str,
    value: Union[PropertyElement, str, Iterable[Union[PropertyElement, str]]],
    calling_resource: str = "",
) -> etree._Element:
    """
    Make an `<uri-prop>` from one or more URIs. The URI(s) can be provided as string or as PropertyElement with a string
    inside. If provided as string, the permissions default to "open".

    Args:
        name: the name of this property as defined in the onto
        value: one or more URIs, as string/PropertyElement, or as iterable of strings/PropertyElements
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        If one of the values is not a valid URI

    Returns:
        an etree._Element that can be appended to the parent resource with resource.append(make_*_prop(...))

    Examples:
        >>> excel2xml.make_uri_prop(":testproperty", "www.test.com")
                <uri-prop name=":testproperty">
                    <uri permissions="open">www.test.com</uri>
                </uri-prop>
        >>> excel2xml.make_uri_prop(":testproperty", excel2xml.PropertyElement("www.test.com", permissions="restricted", comment="example"))
                <uri-prop name=":testproperty">
                    <uri permissions="restricted" comment="example">www.test.com</uri>
                </uri-prop>
        >>> excel2xml.make_uri_prop(":testproperty", ["www.1.com", "www.2.com"])
                <uri-prop name=":testproperty">
                    <uri permissions="open">www.1.com</uri>
                    <uri permissions="open">www.2.com</uri>
                </uri-prop>

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#uri-prop
    """

    # check the input: prepare a list with valid values
    values = prepare_value(value)

    # check value type
    for val in values:
        if not is_uri(str(val.value)):
            msg = (
                f"Failed validation in resource '{calling_resource}', property '{name}': "
                f"'{val.value}' is not a valid URI."
            )
            warnings.warn(DspToolsUserWarning(msg))

    # make xml structure of the valid values
    prop_ = etree.Element(
        "{%s}uri-prop" % xml_namespace_map[None],
        name=name,
        nsmap=xml_namespace_map,
    )
    for val in values:
        kwargs = {"permissions": val.permissions}
        if val.comment and check_notna(val.comment):
            kwargs["comment"] = val.comment
        value_ = etree.Element(
            "{%s}uri" % xml_namespace_map[None],
            **kwargs,  # type: ignore[arg-type]
            nsmap=xml_namespace_map,
        )
        value_.text = str(val.value)
        prop_.append(value_)

    return prop_


def make_region(  # noqa: D417 (undocumented-param)
    label: str,
    id: str,
    permissions: str = "open",
    ark: Optional[str] = None,
    iri: Optional[str] = None,
    creation_date: Optional[str] = None,
) -> etree._Element:
    """
    Creates an empty region element, with the attributes as specified by the arguments.

    Args:
        The arguments correspond 1:1 to the attributes of the `<region>` element.

    Raises:
        Warning: if both an ARK and an IRI are provided
        BaseError: if the creation date is invalid

    Returns:
        The region element, without any children, but with the attributes
        `<region label=label id=id permissions=permissions ark=ark iri=iri></region>`

    Examples:
        >>> region = excel2xml.make_region("label", "id")
        >>> region.append(excel2xml.make_text_prop("hasComment", "This is a comment"))
        >>> region.append(excel2xml.make_color_prop("hasColor", "#5d1f1e"))
        >>> region.append(excel2xml.make_resptr_prop("isRegionOf", "image_0"))
        >>> region.append(excel2xml.make_geometry_prop("hasGeometry", "{...}"))
        >>> root.append(region)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#region
    """

    kwargs = {"label": label, "id": id, "permissions": permissions, "nsmap": xml_namespace_map}
    if ark:
        kwargs["ark"] = ark
    if iri:
        kwargs["iri"] = iri
    if ark and iri:
        msg = f"Both ARK and IRI were provided for resource '{label}' ({id}). The ARK will override the IRI."
        warnings.warn(DspToolsUserWarning(msg))
    if creation_date:
        try:
            DateTimeStamp(creation_date)
        except BaseError:
            raise BaseError(
                f"The region '{label}' (ID: {id}) has an invalid creation date '{creation_date}'. "
                f"Did you perhaps forget the timezone?"
            ) from None
        kwargs["creation_date"] = creation_date

    return etree.Element(
        "{%s}region" % xml_namespace_map[None],
        **kwargs,  # type: ignore[arg-type]
    )


def make_link(  # noqa: D417 (undocumented-param)
    label: str,
    id: str,
    permissions: str = "open",
    ark: Optional[str] = None,
    iri: Optional[str] = None,
    creation_date: Optional[str] = None,
) -> etree._Element:
    """
    Creates an empty link element, with the attributes as specified by the arguments.

    Args:
        The arguments correspond 1:1 to the attributes of the `<link>` element.

    Raises:
        Warning: if both an ARK and an IRI are provided
        BaseError: if the creation date is invalid

    Returns:
        The link element, without any children, but with the attributes
        `<link label=label id=id permissions=permissions ark=ark iri=iri></link>`

    Examples:
        >>> link = excel2xml.make_link("label", "id")
        >>> link.append(excel2xml.make_text_prop("hasComment", "This is a comment"))
        >>> link.append(excel2xml.make_resptr_prop("hasLinkTo", ["resource_0", "resource_1"]))
        >>> root.append(link)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#link
    """

    kwargs = {"label": label, "id": id, "permissions": permissions, "nsmap": xml_namespace_map}
    if ark:
        kwargs["ark"] = ark
    if iri:
        kwargs["iri"] = iri
    if ark and iri:
        msg = f"Both ARK and IRI were provided for resource '{label}' ({id}). The ARK will override the IRI."
        warnings.warn(DspToolsUserWarning(msg))
    if creation_date:
        try:
            DateTimeStamp(creation_date)
        except BaseError:
            raise BaseError(
                f"The link '{label}' (ID: {id}) has an invalid creation date '{creation_date}'. "
                f"Did you perhaps forget the timezone?"
            ) from None
        kwargs["creation_date"] = creation_date

    return etree.Element(
        "{%s}link" % xml_namespace_map[None],
        **kwargs,  # type: ignore[arg-type]
    )


def make_audio_segment(  # noqa: D417 (undocumented-param)
    label: str,
    id: str,
    permissions: str = "open",
) -> etree._Element:
    """
    Creates an empty `<audio-segment>` element, with the attributes as specified by the arguments.

    Args:
        The arguments correspond 1:1 to the attributes of the `<audio-segment>` element.

    Returns:
        The audio-segment element, without any children, but with the attributes
        `<audio-segment label=label id=id permissions=permissions></audio-segment>`

    Examples:
        >>> audio_segment = excel2xml.make_audio_segment("label", "id")
        >>> audio_segment.append(excel2xml.make_isSegmentOf_prop("audio_resource_id"))
        >>> audio_segment.append(excel2xml.make_hasSegmentBounds_prop(segment_start=60, segment_end=120)
        >>> root.append(audio_segment)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#video-segment-audio-segment
    """
    return etree.Element(
        "{%s}audio-segment" % xml_namespace_map[None],
        label=label,
        id=id,
        permissions=permissions,
        nsmap=xml_namespace_map,
    )


def make_video_segment(  # noqa: D417 (undocumented-param)
    label: str,
    id: str,
    permissions: str = "open",
) -> etree._Element:
    """
    Creates an empty `<video-segment>` element, with the attributes as specified by the arguments.

    Args:
        The arguments correspond 1:1 to the attributes of the `<video-segment>` element.

    Returns:
        The video-segment element, without any children, but with the attributes
        `<video-segment label=label id=id permissions=permissions></video-segment>`

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_isSegmentOf_prop("video_resource_id"))
        >>> video_segment.append(excel2xml.make_hasSegmentBounds_prop(segment_start=60, segment_end=120)
        >>> root.append(video_segment)

    See https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#video-segment-audio-segment
    """
    return etree.Element(
        "{%s}video-segment" % xml_namespace_map[None],
        label=label,
        id=id,
        permissions=permissions,
        nsmap=xml_namespace_map,
    )


def make_isSegmentOf_prop(
    target_id: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<isSegmentOf>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        target_id: ID of target video/audio resource
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the target_id is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_isSegmentOf_prop("video_resource_id"))
        >>> video_segment.append(excel2xml.make_hasSegmentBounds_prop(segment_start=60, segment_end=120)
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video segment with `segment.append(make_isSegmentOf_prop(...))`
    """
    if not isinstance(target_id, str) or not check_notna(target_id):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'isSegmentOf': "
            f"The following doesn't seem to be a valid ID of a target resource: '{target_id}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}isSegmentOf" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop.text = target_id
    return prop


def make_relatesTo_prop(
    target_id: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<relatesTo>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        target_id: ID of the related resource
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the target_id is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_relatesTo_prop("resource_id"))
        >>> video_segment.append(excel2xml.make_relatesTo_prop("other_resource_id")
        >>> # add other properties
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video segment with `segment.append(make_relatesTo_prop(...))`
    """
    if not isinstance(target_id, str) or not check_notna(target_id):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'relatesTo': "
            f"The following doesn't seem to be a valid ID of a target resource: '{target_id}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}relatesTo" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop.text = target_id
    return prop


def make_hasSegmentBounds_prop(
    segment_start: int | float,
    segment_end: int | float,
    permissions: str = "open",
    comment: str | None = None,
    calling_resource: str = "",
) -> etree._Element:
    """
    Make a `<hasSegmentBounds>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        segment_start: start, in seconds, counted from the beginning of the audio/video
        segment_end: end, in seconds, counted from the beginning of the audio/video
        permissions: Defaults to "open".
        comment: Optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If start or end are not integers or floats
        - If start is bigger than end

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_isSegmentOf_prop("video_resource_id"))
        >>> video_segment.append(excel2xml.make_hasSegmentBounds_prop(segment_start=60, segment_end=120)
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video segment with `segment.append(make_hasSegmentBounds_prop(...))`
    """
    if not isinstance(segment_start, (int, float)) or not isinstance(segment_end, (int, float)):
        try:
            segment_start = float(segment_start)
            segment_end = float(segment_end)
        except ValueError:
            msg = (
                f"Validation Error in resource '{calling_resource}', property 'hasSegmentBounds': "
                f"The start and the end of an audio/video segment must be integers or floats, "
                f"but you provided: {segment_start=} and {segment_end=}"
            )
            warnings.warn(DspToolsUserWarning(msg))
    if (
        isinstance(segment_start, (int, float))
        and isinstance(segment_end, (int, float))
        and segment_start > segment_end
    ):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'hasSegmentBounds': "
            f"The start of an audio/video segment must be less than the end, "
            f"but you provided: {segment_start=} and {segment_end=}"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element(
        "{%s}hasSegmentBounds" % xml_namespace_map[None],
        segment_start=str(segment_start),
        segment_end=str(segment_end),
        permissions=permissions,
    )
    if comment:
        prop.set("comment", comment)
    return prop


def make_hasTitle_prop(
    title: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<hasTitle>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        title: the title of the segment
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the title is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_hasTitle_prop("title of my segment"))
        >>> # add other properties
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video resource with `segment.append(make_hasTitle_prop(...))`
    """
    if not isinstance(title, str) or not check_notna(title):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'hasTitle': "
            f"The following doesn't seem to be a valid string: '{title}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}hasTitle" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop.text = title
    return prop


def make_hasKeyword_prop(
    keyword: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<hasKeyword>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        keyword: a keyword of the segment
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the keyword is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_hasKeyword_prop("keyword of my segment"))
        >>> video_segment.append(excel2xml.make_hasKeyword_prop("another keyword"))
        >>> # add other properties
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video resource with `segment.append(make_hasKeyword_prop(...))`
    """
    if not isinstance(keyword, str) or not check_notna(keyword):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'hasKeyword': "
            f"The following doesn't seem to be a valid string: '{keyword}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}hasKeyword" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop.text = keyword
    return prop


def make_hasComment_prop(
    comment_text: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<hasComment>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        comment_text: a text with some background info about the segment. Can be formatted with tags.
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the comment text is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_hasComment_prop("comment of my segment"))
        >>> video_segment.append(excel2xml.make_hasComment_prop("another comment"))
        >>> # add other properties
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video resource with `segment.append(make_hasComment_prop(...))`
    """
    if not isinstance(comment_text, str) or not check_notna(comment_text):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'hasComment': "
            f"The following doesn't seem to be a valid string: '{comment_text}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}hasComment" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop = _add_richtext_to_etree_element(comment_text, prop)
    return prop


def make_hasDescription_prop(
    description: str, permissions: str = "open", comment: str | None = None, calling_resource: str = ""
) -> etree._Element:
    """
    Make a `<hasDescription>` property for a `<video-segment>` or `<audio-segment>`.

    Args:
        description: a text with some background info about the segment. Can be formatted with tags.
        permissions: defaults to "open".
        comment: optional comment for this property. Defaults to None.
        calling_resource: the name of the parent resource (for better error messages)

    Warns:
        - If the description is not a valid string

    Examples:
        >>> video_segment = excel2xml.make_video_segment("label", "id")
        >>> video_segment.append(excel2xml.make_hasDescription_prop("a description of my segment"))
        >>> video_segment.append(excel2xml.make_hasDescription_prop("another description"))
        >>> # add other properties
        >>> root.append(video_segment)

    Returns:
        an etree._Element that can be appended to an audio/video resource with `segment.append(make_hasDescription_prop(...))`
    """
    if not isinstance(description, str) or not check_notna(description):
        msg = (
            f"Validation Error in resource '{calling_resource}', property 'hasDescription': "
            f"The following doesn't seem to be a valid string: '{description}'"
        )
        warnings.warn(DspToolsUserWarning(msg))
    prop = etree.Element("{%s}hasDescription" % xml_namespace_map[None], permissions=permissions)
    if comment:
        prop.set("comment", comment)
    prop = _add_richtext_to_etree_element(description, prop)
    return prop


def create_json_excel_list_mapping(
    path_to_json: str,
    list_name: str,
    excel_values: Iterable[str],
    sep: str = '+"*ç%&/()=',
    corrections: Optional[dict[str, str]] = None,
) -> dict[str, str]:
    """
    Often, data sources contain list values that aren't identical to the name of the node in the list of the JSON
    project file (colloquially: ontology). In order to create a correct XML for the `dsp-tools xmlupload`, a mapping is
    necessary. This function takes a JSON list and an Excel column containing list-values, and tries to match them
    automatically based on similarity. The result is a dict of the form {excel_value: list_node_name}.

    Alternatively, consider using the function create_json_list_mapping(), which also builds a dictionary,
    but from the names and labels in the JSON list, which is less error-prone than this function's approach. However,
    this function has the advantage that it even works when your data source doesn't use the list labels correctly.

    Args:
        path_to_json: path to the JSON project file
        list_name: name of the list in the JSON project file (can also be a nested list)
        excel_values: the Excel column (e.g. as list) with the list values in it
        sep: separator string, if the cells in the Excel contain more than one list entry
        corrections: dict with wrong entries, each pointing to its correct counterpart

    Raises:
        Warning: if there is an Excel value that couldn't be matched
        Exception: if the path doesn't point to a JSON project file

    Returns:
        dict of the form ``{excel_value: list_node_name}``.
            Every excel_value is stripped, and also present in a lowercase form.

    Examples:
        >>> json_list_nodes = [
                {
                    "name": "giraffe",
                    "labels": {"en": "giraffe"}
                },
                {
                    "name": "antelope",
                    "labels": {"en": "antelope"}
                }
            ]
        >>> excel_row_1 = ["Giraffeeh ", " Antiloupe", "Girraffe , Antiloupe "]
        >>> json_excel_list_mapping = {
                "Giraffeeh": "giraffe",
                "giraffeeh": "giraffe",
                "Girraffe": "giraffe",
                "girraffe": "giraffe",
                "Antiloupe": "antelope",
                "antiloupe": "antelope"
            }
    """

    # avoid mutable default argument
    corrections = corrections or {}

    # split the values, if necessary
    excel_values_new = []
    for val in excel_values:
        if isinstance(val, str):
            excel_values_new.extend([x.strip() for x in val.split(sep) if x])

    # read the list of the JSON project (works also for nested lists)
    with open(path_to_json, encoding="utf-8") as f:
        json_file = json.load(f)
    json_subset = []
    for elem in json_file["project"]["lists"]:
        if elem["name"] == list_name:
            json_subset = elem["nodes"]
    json_values = set(_nested_dict_values_iterator(json_subset))

    # build dictionary with the mapping, based on string similarity
    res = {}
    for excel_value in excel_values_new:
        excel_value_corrected = corrections.get(excel_value, excel_value)
        excel_value_simpl = simplify_name(excel_value_corrected)  # increase match probability by removing illegal chars
        if matches := difflib.get_close_matches(
            word=excel_value_simpl,
            possibilities=json_values,
            n=1,
            cutoff=0.6,
        ):
            res[excel_value] = matches[0]
            res[excel_value.lower()] = matches[0]
        else:
            msg = (
                f"Did not find a close match to the excel list entry '{excel_value}' "
                f"among the values in the JSON project list '{list_name}'"
            )
            warnings.warn(DspToolsUserWarning(msg))

    return res


def _nested_dict_values_iterator(dicts: list[dict[str, Any]]) -> Iterable[str]:
    """
    Yield all values of a nested dictionary.

    Args:
        dicts: list of nested dictionaries

    Yields:
        values of the nested dictionaries
    """
    # Credits: https://thispointer.com/python-iterate-loop-over-all-nested-dictionary-values/
    for _dict in dicts:
        if "nodes" in _dict:
            yield from _nested_dict_values_iterator(_dict["nodes"])
        if "name" in _dict:
            yield _dict["name"]


def create_json_list_mapping(
    path_to_json: str,
    list_name: str,
    language_label: str,
) -> dict[str, str]:
    """
    Often, data sources contain list values named after the "label" of the JSON project list node, instead of the "name"
    which is needed for the `dsp-tools xmlupload`. In order to create a correct XML, you need a dictionary that maps the
    "labels" to their correct "names".

    Alternatively, consider using the method create_json_excel_list_mapping(), which also creates a dictionary, but maps
    values from your data source to list node names from the JSON project file, based on similarity.

    Args:
        path_to_json: path to a JSON project file (a.k.a. ontology)
        list_name: name of a list in the JSON project (works also for nested lists)
        language_label: which language of the label to choose

    Returns:
        a dictionary of the form {label: name}
    """
    with open(path_to_json, encoding="utf-8") as f:
        json_file = json.load(f)
    json_subset = [x for x in json_file["project"]["lists"] if x["name"] == list_name]
    # json_subset is a list containing one item, namely the json object containing the entire json-list

    res = {}
    for label, name in _name_label_mapper_iterator(json_subset, language_label):
        if name != list_name:
            res[label] = name
            res[label.strip().lower()] = name

    return res


def _name_label_mapper_iterator(
    json_subset: list[dict[str, Any]],
    language_label: str,
) -> Iterable[tuple[str, str]]:
    """
    Go through list nodes of a JSON project and yield (label, name) pairs.

    Args:
        json_subset: list of DSP lists (a DSP list being a dictionary with the keys "name", "labels" and "nodes")
        language_label: which language of the label to choose

    Yields:
        (label, name) pairs
    """
    for node in json_subset:
        # node is the json object containing the entire json-list
        if "nodes" in node:
            # "nodes" is the json sub-object containing the entries of the json-list
            yield from _name_label_mapper_iterator(node["nodes"], language_label)
            # each yielded value is a (label, name) pair of a single list entry
        if "name" in node:
            yield (node["labels"][language_label], node["name"])
            # the actual values of the name and the label


def write_xml(
    root: etree._Element,
    filepath: str | Path,
) -> None:
    """
    Write the finished XML to a file.

    Args:
        root: etree Element with the entire XML document
        filepath: where to save the file

    Warning:
        if the XML is not valid according to the schema
    """
    etree.indent(root, space="    ")
    xml_string = etree.tostring(
        root,
        encoding="unicode",
        pretty_print=True,
        doctype='<?xml version="1.0" encoding="UTF-8"?>',
    )
    xml_string = xml_string.replace(r"\'", "'")
    with open(filepath, "w", encoding="utf-8") as f:
        f.write(xml_string)
    try:
        validate_xml_file(input_file=filepath)
        print(f"The XML file was successfully saved to {filepath}")
    except BaseError as err:
        msg = (
            f"The XML file was successfully saved to {filepath}, "
            f"but the following Schema validation error(s) occurred: {err.message}"
        )
        warnings.warn(DspToolsUserWarning(msg))



from __future__ import annotations

from pathlib import Path

import pytest
import regex
from lxml import etree

from dsp_tools.utils.xml_utils import parse_and_clean_xml_file


@pytest.fixture
def data_systematic_unclean() -> Path:
    return Path("testdata/xml-data/test-data-systematic.xml")


@pytest.fixture
def data_systematic_cleaned() -> etree._Element:
    return parse_and_clean_xml_file(Path("testdata/xml-data/test-data-systematic.xml"))


def clean_resulting_tree(tree: etree._Element) -> str:
    cleaned_str = regex.sub("\n", "", etree.tostring(tree, encoding=str))
    return regex.sub(" +", " ", cleaned_str)


def test_parse_and_clean_xml_file_same_regardless_of_input(data_systematic_unclean: Path) -> None:
    from_tree = parse_and_clean_xml_file(data_systematic_unclean)
    from_file = parse_and_clean_xml_file(Path("testdata/xml-data/test-data-systematic.xml"))
    cleaned_from_file = clean_resulting_tree(from_file)
    cleaned_from_tree = clean_resulting_tree(from_tree)
    assert (
        cleaned_from_file == cleaned_from_tree
    ), "The output must be equal, regardless if the input is a path or parsed."


def test_regions_links_before(data_systematic_unclean: Path) -> None:
    data = etree.parse(data_systematic_unclean)
    regions_links_before = [e for e in data.iter() if regex.search("region|link", str(e.tag))]
    assert len(regions_links_before) == 5


def test_regions_links_after(data_systematic_cleaned: etree._Element) -> None:
    regions_links_after = [e for e in data_systematic_cleaned.iter() if regex.search("region|link", str(e.tag))]
    assert len(regions_links_after) == 0, (
        "The tags <region> and <link> must be transformed "
        'to their technically correct form <resource restype="Region/LinkObj">'
    )


def test_comment_removal(data_systematic_unclean: Path, data_systematic_cleaned: etree._Element) -> None:
    data = etree.parse(data_systematic_unclean)
    comments_before = [e for e in data.iter() if isinstance(e, etree._Comment)]
    assert len(comments_before) > 0
    comments_after = [e for e in data_systematic_cleaned.iter() if isinstance(e, etree._Comment)]
    assert len(comments_after) == 0, (
        "properties that are commented out would break the the constructor of the class XMLProperty, "
        "if they are not removed in the parsing process"
    )


if __name__ == "__main__":
    pytest.main([__file__])
